#include "navigation_pile.hpp"
#include <cmath> // defines ilogb
#include <vector> // defines std::vector
#include <algorithm> // defines std::copy

#include "dsexceptions.h" // defines Overflow and Underflow
#include "debugging.c++" // defines show


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::offset
navigation_pile<T, Container, Compare>::get(
  navigation_pile<T, Container, Compare>::bit_index i,
  navigation_pile<T, Container, Compare>::level gamma
) const {
  return cphstl::get(i, gamma);
}


template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::set(
  navigation_pile<T, Container, Compare>::bit_index i,
  navigation_pile<T, Container, Compare>::level gamma,
  navigation_pile<T, Container, Compare>::offset lambda
) {
  cphstl::set(i, gamma, lambda);
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::level 
navigation_pile<T, Container, Compare>::depth(
navigation_pile<T, Container, Compare>::node_index i
) const {
  return ilogb(1 + i);
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::level 
navigation_pile<T, Container, Compare>::height(
navigation_pile<T, Container, Compare>::node_index i
) const {
    return _eta - depth(i);
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::first_child(
navigation_pile<T, Container, Compare>::node_index i
) const {
  return (i << 1) + 1;
  //return 2 * i + 1;
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::second_child(
navigation_pile<T, Container, Compare>::node_index i
) const {
  return (i << 1) + 2;
  //return 2 * i + 2;
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::parent(
navigation_pile<T, Container, Compare>::node_index i
) const {
  return (i - 1) >> 1;
  //return (i - 1) / 2; parent(0) == 0 not parent(0) == -1
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::ancestor(
navigation_pile<T, Container, Compare>::node_index i, 
navigation_pile<T, Container, Compare>::level delta) const {
  return ((i + 1) >> delta) - 1;
  //return (i + 1) / pow_two(delta) - 1;
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::root() const {
  level delta;
  if (_n <= 1) {
    delta = _eta - 1;
  }
  else {
    delta = _eta - ceiling_log(_n);
  }

  return pow_two(delta) - 1;
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::leaf_index 
navigation_pile<T, Container, Compare>::last_leaf() const{
  //return pow_two(_eta) + _n - 2;
  return _first_leaf + _n - 1;
}


template <class T, class Container, class Compare>
bool
navigation_pile<T, Container, Compare>::empty() const {
 return (m == 0);
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::size_type
navigation_pile<T, Container, Compare>::size() const {
  return _n;
}


template <class T, class Container, class Compare>
bool 
navigation_pile<T, Container, Compare>::is_first_child(
navigation_pile<T, Container, Compare>::node_index i
) const {
  return (i & 1 == 1);
}


template <class T, class Container, class Compare>
bool 
navigation_pile<T, Container, Compare>::is_root(
navigation_pile<T, Container, Compare>::node_index i
) const {
  return i == _root;
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::bit_index
navigation_pile<T, Container, Compare>::start_of_offset(
navigation_pile<T, Container, Compare>::branch_index i,
navigation_pile<T, Container, Compare>::level gamma
) const {
  return (-_eta + (pow_two(_eta - gamma) << 1) - 2 + i * gamma + gamma);
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::leaf_index 
navigation_pile<T, Container, Compare>::element_to_leaf(
navigation_pile<T, Container, Compare>::element_index ell
) const {
return _first_leaf + ell;
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::element_index 
navigation_pile<T, Container, Compare>::leaf_to_element(
navigation_pile<T, Container, Compare>::leaf_index L
) const {
  return L - _first_leaf;
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::element_index
navigation_pile<T, Container, Compare>::jump_to_element(
  navigation_pile<T, Container, Compare>::branch_index i,
  navigation_pile<T, Container, Compare>::offset &lambda
) const {
  level gamma = height(i);
  bit_index s = start_of_offset(i, gamma);
  lambda = B.get(s, gamma);
  level pow_two_gamma = pow_two(gamma);
  leaf_index start_of_leaf_sequence = pow_two_gamma * i + pow_two_gamma -  1;
  element_index ell = leaf_to_element(start_of_leaf_sequence + lambda);
  return ell;
}

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::element_index
navigation_pile<T, Container, Compare>::jump_to_element(
  navigation_pile<T, Container, Compare>::branch_index i
) const {
  level gamma = height(i);
  bit_index s = start_of_offset(i, gamma);
  offset lambda = B.get(s, gamma);
  level pow_two_gamma = pow_two(gamma);
  leaf_index start_of_leaf_sequence = pow_two_gamma * i + pow_two_gamma -  1;
  element_index ell = leaf_to_element(start_of_leaf_sequence + lambda);
  return ell;
}


template <class T, class Container, class Compare>
navigation_pile<T, Container, Compare>::navigation_pile(size_type N, 
							const Compare& x, 
							const Container& Y
							) : A(Y), B(int(pow_two(ilogb_zero(N - 1) + 2)), false), _n(Y.size()), _N(N), _eta(ilogb_zero(N - 1) + 1), _first_leaf(pow_two(_eta) - 1), _root(root()), _root_n(pow_two(height(_root) - 1)) {
  less = x;
  if (_n <= 1) return;
  make_pile();
}


template <class T, class Container, class Compare>
template <class InputIterator>
navigation_pile<T, Container, Compare>:: navigation_pile(InputIterator first, 
							 InputIterator last, 
							 size_type N,
							 const Compare& x,
							 const Container& Y
							 ) : A(Y), B(int(pow_two(ilogb_zero(N - 1) + 2)), false), _n((last - first) + Y.size()), _N(N), _eta(ilogb_zero(N - 1) + 1), _first_leaf(pow_two(_eta) - 1), _root(root()), _root_n(pow_two(height(_root) - 1)) {
  A.insert(A.end(), first, last);
  less = x;
  if (_n <= 1) return;
  make_pile();
}


template <class T, class Container, class Compare>
navigation_pile<T, Container, Compare>::~navigation_pile(){
}


template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::make_pile() {
  branch_index i;
  element_index m;
  element_index ell;
  node_index fc;
  node_index sc;
  bit_index s; 
  offset lambda_ell; 
  offset lambda_m; 
  level gamma;
  node_index last_node_in_use;
  branch_index f = parent(_first_leaf);
  for (int e = int(parent(last_leaf())); e >= int(f); --e ) {
    handle_height_one_branch(e);
    last_node_in_use = last_leaf();
    i = e;
    while (is_first_child(i) && !is_root(i)) {
      i = parent(i);
      fc = first_child(i);
      sc = second_child(i);
      last_node_in_use = parent(last_node_in_use);
      gamma = height(i);
      s = start_of_offset(i, gamma);
      ell = jump_to_element(fc, lambda_ell);
      if (sc <= last_node_in_use) {
	m = jump_to_element(sc, lambda_m);
	less(A[m], A[ell]) ? B.set(s, gamma, lambda_ell) : B.set(s, gamma, pow_two(gamma - 1) + lambda_m);
      }
      else { 
	B.set(s, gamma, lambda_ell);
      }
    }
  }
}


template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::handle_height_one_branch(
navigation_pile<T, Container, Compare>::branch_index i
) {
  bit_index s = start_of_offset(i, 1);
  node_index fc = first_child(i);
  node_index sc = fc + 1;
  element_index ell = leaf_to_element(fc);
  if (sc <= last_leaf()) {
    element_index m = leaf_to_element(sc);

    bool v = less(A[m],A[ell]) ? 0 : 1;
    B.set(s, 1, v);
  }
  else {
    B.set(s, 1, 0);
  }
}


template <class T, class Container, class Compare>
const typename navigation_pile<T, Container, Compare>::value_type& 
navigation_pile<T, Container, Compare>::top() const {
  return A[jump_to_element(_root)];
}


template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::push(
const navigation_pile<T, Container, Compare>::value_type& x
) {
  if(_n >= _N) throw Overflow();
  level old_root = _root;
  insert_new_element(x);
  if(_n - 1 == _root_n << 1) {
    _root_n = _root_n << 1;
    _root = parent(_root);
    level gamma = height(old_root);
    bit_index s = start_of_offset(old_root, gamma);
    offset lambda = B.get(s, gamma);
    s = start_of_offset(_root, gamma + 1);
    B.set(s, gamma + 1, lambda);
  }
  node_index j = last_leaf();
  if ((j & 1)) {
    j = update_partial_path(j);
  }
  node_index i = binary_search_on_special_path(j, x);
  
  // Alternativ to binary search
  //node_index i = bottom_up_on_special_path(j, x);
  
  update_partial_path(j, i, last_leaf());

}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::binary_search_on_special_path(
navigation_pile<T, Container, Compare>::node_index j, 
const navigation_pile<T, Container, Compare>::element& x
) {
  level delta = depth(j) - depth(_root);
  branch_index i;
  level half;
  element_index ell;
  while (delta > 0) {
    half = delta / 2;
    i = ancestor(j, delta - half);
    ell = jump_to_element(i);
    if (less(A[ell], x)) {
      j = i;
      delta = half;
    }
    else {
      delta = delta - half -  1;
    }
  }
  return j;
}


template <class T, class Container, class Compare>
void  
navigation_pile<T, Container, Compare>::update_partial_path(
navigation_pile<T, Container, Compare>::node_index j, 
navigation_pile<T, Container, Compare>::node_index i, 
navigation_pile<T, Container, Compare>::leaf_index K
) {
  bit_index s;
  node_index pow_two_gamma;
  leaf_index start_of_leaf_sequence;
  offset lambda;
  level gamma = height(j);
  while (j != i) {
    j = parent(j);
    gamma = gamma + 1;
    s = start_of_offset(j, gamma);
    pow_two_gamma = pow_two(gamma);
    start_of_leaf_sequence = pow_two_gamma * j + pow_two_gamma -  1;
    lambda = K - start_of_leaf_sequence;
    B.set(s, gamma, lambda);
  }
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::update_partial_path(
navigation_pile<T, Container, Compare>::leaf_index K
) {
  bit_index s;
  node_index pow_two_gamma;
  leaf_index start_of_leaf_sequence;
  offset lambda;
  node_index j = K;
  node_index i = parent(j);
  node_index rootnode = _root;
  size_type le = _n - 1;
  level gamma = height(j);
  while (i != rootnode && ((le & 1) == 0)) {
    le = le >> 1;
    s = start_of_offset(j, gamma);
    gamma = gamma + 1;
    pow_two_gamma = pow_two(gamma);
    start_of_leaf_sequence = pow_two_gamma * i + pow_two_gamma -  1;
    lambda = K - start_of_leaf_sequence;
    B.set(s, gamma, lambda);
    j = i;
    i = parent(i);
  }
  return j;
}


// Pop using first ansector technique
template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::pop() {
  if(_n == 0) throw Underflow();
  if (_n == 1) {
    erase_last_element();
    return;
  }
  leaf_index ll = last_leaf();
  element_index k;
  element_index m = jump_to_element(_root);
  const element_index ell = leaf_to_element(ll);
  branch_index i;
  i = first_ancestor_with_two_children(ll);  
  if (height(i) == 1) {
    k = ell - 1;
  }
  else {
    k = jump_to_element(first_child(i));
  }
  bool new_root = (_n - 1 == _root_n && _n - 1 != 1);
  
  if (ell == m) {
    erase_last_element();
    if (new_root) {
      _root_n = _root_n >> 1;
      _root = first_child(_root);
    }
    update_full_path(last_leaf());
  }
  else if (jump_to_element(i) == ell) {
    A[m] = A[k];
    A[k] = A[ell];
    erase_last_element();
    if (new_root) {
      _root_n = _root_n >> 1;
      _root = first_child(_root);
    }
    leaf_index lk = element_to_leaf(k);
    branch_index plk = parent(lk);
    bit_index s;
    offset kappa;
    node_index pow_two_gamma;
    leaf_index start_of_leaf_sequence;
    offset mu;
    level gamma = height(plk);
    s = start_of_offset(plk, gamma);
    kappa = B.get(s, gamma);
    pow_two_gamma = pow_two(gamma);
    start_of_leaf_sequence = pow_two_gamma * plk + pow_two_gamma - 1;
    if (kappa == ll) {
    mu = lk - start_of_leaf_sequence;
    B.set(s, gamma, mu);
    }
    update_partial_path(i, _root, lk, ll);
    update_full_path(element_to_leaf(m));
  }
  else {
    A[m] = A[ell];
    erase_last_element();
    if (new_root) {
      _root_n = _root_n >> 1;
      _root = first_child(_root);
    }
    update_full_path(element_to_leaf(m));
  }
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::first_ancestor_with_two_children(
navigation_pile<T, Container, Compare>::leaf_index L
) {
  size_type le = _n - 1;
  unsigned int i = 1;

  while ((le & 1) == 0) {
    i++;
    le = le >> 1;
  }
  return (ancestor(L, i));
}


template <class T, class Container, class Compare>
void navigation_pile<T, Container, Compare>::first_two_ancestors_with_more_than_one_child(
  navigation_pile<T, Container, Compare>::branch_index &i_1, 
  navigation_pile<T, Container, Compare>::branch_index &i_2,
  navigation_pile<T, Container, Compare>::leaf_index L
) {
  size_type le = _n - 1;
  unsigned int i = 1;

  while ((le & 1) == 0) {
    i++;
    le = le >> 1;
  }
  i_1 = ancestor(L, i);
  if (is_root(i_1)) {
    i_2 = i_1;
    return;
  }
  i = i + 1;
  le = le >> 1;
  while ((le & 1) == 0) {
    i++;
    le = le >> 1;
  }
  i_2 = ancestor(L, i);
}


template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::update_full_path(
navigation_pile<T, Container, Compare>::leaf_index L
) {
  node_index last_node_in_use = last_leaf();
  branch_index i = parent(L);
  handle_height_one_branch(i);
  level gamma = 1;
  element_index two_to_gamma_minus_one = 1;
  element_index start_of_element_sequence = first_child(i) - _first_leaf;
  bit_index s_i = start_of_offset(i, gamma);
  bit_index s_fc;
  offset lambda_ell = B.get(s_i, gamma);
  offset lambda_m; 
  while (!is_root(i)) {
    if (i & 1) {
      s_fc = s_i;
      i = parent(i);
      last_node_in_use = parent(last_node_in_use);
      gamma = gamma + 1;
      two_to_gamma_minus_one = two_to_gamma_minus_one << 1;
      s_i = start_of_offset(i, gamma);
      if (second_child(i) <= last_node_in_use) {
	lambda_m = B.get(s_fc + gamma - 1, gamma - 1);
	if (less(A[start_of_element_sequence + two_to_gamma_minus_one + lambda_m], A[start_of_element_sequence + lambda_ell])) {
	  B.set(s_i, gamma, lambda_ell);
	}
	else {
	  lambda_ell = two_to_gamma_minus_one + lambda_m;
	  B.set(s_i, gamma, lambda_ell);
	}
      }
      else { 
	B.set(s_i, gamma, lambda_ell);      
      }
    }
    else {
      two_to_gamma_minus_one = two_to_gamma_minus_one << 1;
      s_fc = s_i - gamma;
      start_of_element_sequence = start_of_element_sequence - two_to_gamma_minus_one;
      i = parent(i);
      last_node_in_use = parent(last_node_in_use);
      lambda_m = B.get(s_fc, gamma);
      gamma = gamma + 1;
      s_i = start_of_offset(i, gamma);
      if (less(A[start_of_element_sequence + two_to_gamma_minus_one + lambda_ell], A[start_of_element_sequence + lambda_m])) {
	B.set(s_i, gamma, lambda_m);
	lambda_ell = lambda_m;
      }
      else {
	lambda_ell = two_to_gamma_minus_one + lambda_ell;
	B.set(s_i, gamma, lambda_ell);
      }
    }
  }
}


template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::update_partial_path(
navigation_pile<T, Container, Compare>::branch_index j,
navigation_pile<T, Container, Compare>::branch_index i, 
navigation_pile<T, Container, Compare>::leaf_index K, 
navigation_pile<T, Container, Compare>::leaf_index L
) {
  bit_index s;
  offset kappa;
  node_index pow_two_gamma;
  leaf_index start_of_leaf_sequence;
  offset lambda;
  offset mu;
  level gamma = height(j);
  while (i != j) {
    s = start_of_offset(j, gamma);
    kappa = B.get(s, gamma);
    pow_two_gamma = pow_two(gamma);
    start_of_leaf_sequence = pow_two_gamma * j + pow_two_gamma - 1;
    lambda = L - start_of_leaf_sequence;
    if (kappa != lambda) return;
    mu = K - start_of_leaf_sequence;
    B.set(s, gamma, mu);
    j = parent(j);
    gamma = gamma + 1;
  }
}


template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::insert_new_element(
const navigation_pile<T, Container, Compare>::element& z
) {
  A.push_back(z);
  _n++;
}


template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::erase_last_element() {
  A.pop_back();
  _n--;
}

// Calculates Sibling
template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index
navigation_pile<T, Container, Compare>::sibling(
navigation_pile<T, Container, Compare>::node_index i
) const {
  bool odd = i & 1;
  return (i + odd - !odd);
}


// Calculates power z of two
template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::pow_two(
navigation_pile<T, Container, Compare>::level z
) const {
  return (1 << z);
  //return int(pow(2, z));
}


// Calculates ceils{ log_2 n }
template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::size_type 
navigation_pile<T, Container, Compare>::ceiling_log(
navigation_pile<T, Container, Compare>::size_type n
) const {
  size_type i = ilogb(n);
  return i + (pow_two(i) != n);
}

// Calculates ilogb but returning zero if i <= 0
template <class T, class Container, class Compare>
int 
navigation_pile<T, Container, Compare>::ilogb_zero(
double i
) const {
  if (i <= 0) return 0;
  return ilogb(i);
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::bottom_up_on_special_path(
navigation_pile<T, Container, Compare>::node_index j, 
const navigation_pile<T, Container, Compare>::element& x
) {
  element_index ell;
  branch_index i = parent(j);
  while (i >= _root) {
    ell = jump_to_element(i);
    if (less(A[ell], x)) {
      j = i;
      if (i == 0) return j;  // parent(0) != -1
      i = parent(i);
    }
    else {
      return j;
    }
  }
  return j;   
}


#include <is_sorted.c++> // defines cphstl::is_sorted 

// Utility function used for testing
template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::sort_pile(
size_type n, 
const Compare& x
) {
  less = x;
  Container t(n);
  std::copy(A.begin(), A.end(), t.begin());
  std::sort(t.begin(), t.end(), less);
  Container E(n);
  // use i-- insted of int i = 0 and n++ to make is_sorted work
  int i = n - 1;  
  while(size() > 0) {
    E[i] = top();
    pop();
    i--;
  }
  assert(cphstl::is_sorted(E.begin(), E.end(), less));
  assert(std::equal(E.begin(), E.end(), t.begin()));
}
