#include "navigation_pile.hpp"
#include <cmath> // defines ilogb
#include <vector> // defines std::vector
#include <algorithm> // defines std::copy

#include "dsexceptions.h" // defines Overflow and Underflow
#include "debugging.c++" // defines show

// priority_queue(const input_sequence& X, const ordering& f,  const container_sequence& Y);
// Effect: Construct a priority queue containing the elements
// stored in the sequence referred to by X using the function 
// referred to by f as the ordering function and a copy of the 
// sequence referred to by Y as the underlying container for the elements.
// In particular, note that the elements are to be copied from the input
// sequence to a separate container sequence.

template <class T, class Container, class Compare>
priority_queue<T, Container, Compare>::priority_queue(size_type N,
						      const Compare& x,
						      const Container& Y) {
  navigation_pile<T, Container, Compare>::navigation_pile(N, x, Y);
}


template <class T, class Container, class Compare>
template <typename InputIterator>
priority_queue<T, Container, Compare>::priority_queue(InputIterator first, 
						      InputIterator last,
						      size_type N,
						      const Compare& x,
						      const Container& Y) {
  navigation_pile<T, Container, Compare>::navigation_pile(first, last, N, x, Y);
}


template <class T, class Container, class Compare>
priority_queue<T, Container, Compare>::~priority_queue(){
}


template <class T, class Container, class Compare>
bool
priority_queue<T, Container, Compare>::empty() const {
 return(navigation_pile<T, Container, Compare>::empty()); }


template <class T, class Container, class Compare>
typename priority_queue<T, Container, Compare>::size_type
priority_queue<T, Container, Compare>::size() const { 
  return(navigation_pile<T, Container, Compare>::size()); }


// const element& top() const;
// Effect: Return a reference to an element whose priority is highest
// among the elements stored in the priority queue. We call such an element the
// top element, i.e.for that element y and for every other
// element x in the priority queue, less(y, x) must return false. 

template <class T, class Container, class Compare>
const typename priority_queue<T, Container, Compare>::value_type& 
priority_queue<T, Container, Compare>::top() const {
  return(navigation_pile<T, Container, Compare>::top());
}


// void push(const element& x);
// Effect: Insert a copy of the element referred to by x into the priority queue.

template <class T, class Container, class Compare>
void 
priority_queue<T, Container, Compare>::push(
const priority_queue<T, Container, Compare>::value_type& x) {
  navigation_pile<T, Container, Compare>::push(x);
}


// void pop();
// Requirement: The priority queue should not be empty.
// Effect: Erase the top element from the priority queue.

template <class T, class Container, class Compare>
void 
priority_queue<T, Container, Compare>::pop() {
  navigation_pile<T, Container, Compare>::pop();
}


// According to the C++ standard S 23.2.3 ISO, other operations 
// (like size()) should be supported as well, but normally these are trivial to implement.

//offset get(bit_index i, level gamma) const;

// Requirement: gamma <= w.
// Effect: Return the integer corresponding to the bit sequence B[i.. i + gamma).
// We use the shorthand notation lambda <- B[i.. i + gamma) for this operation.



//void set(bit_index i, level gamma, offset lambda);

// Requirement: gamma <= w.
// Effect: Update the contents of the bit sequence B[i.. i + gamma) using the
// gamma low-order bits of lambda. In brief, we denote this
// operation as B[i.. i + gamma) <- lambda.

// The implementation of these functions is straightforward using O(1) left
// and right shifts, and bitwise boolean instructions.



// The navigation-pile class has several member functions which 
// simplify the description of the other member functions. 
// To make the bounds for the number of element comparisons and machine
// instructions depend on n, not on N, any of the nodes on the left arm of 
// the complete binary tree can be the root.
// Below we give the first collection of member functions.


// level depth(node_index i) const;
// Effect: RETURN floors{log_2(1 + i)};

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::level 
navigation_pile<T, Container, Compare>::depth(
navigation_pile<T, Container, Compare>::node_index i
) const {
  return ilogb(1 + i);
}


// level height(node_index i) const;
// Effect: RETURN eta - depth(i);
// eta <- depth of hole pile structor

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::level 
navigation_pile<T, Container, Compare>::height(
navigation_pile<T, Container, Compare>::node_index i
) const {
    return _eta - depth(i);
}


// node_index first_child(node_index i) const;
// Effect: RETURN 2 * i + 1;

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::first_child(
navigation_pile<T, Container, Compare>::node_index i
) const {
  return (i << 1) + 1;
  //return 2 * i + 1;
}


// node_index second_child(node_index i) const;
// Effect: RETURN 2 * i + 2;

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::second_child(
navigation_pile<T, Container, Compare>::node_index i
) const {
  return (i << 1) + 2;
  //return 2 * i + 2;
}


// node_index parent(node_index i) const;
// Effect: RETURN floors{(i -  1) / 2};

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::parent(
navigation_pile<T, Container, Compare>::node_index i
) const {
  return (i - 1) >> 1;
  //return (i - 1) / 2; parent(0) == 0 not parent(0) == -1
}


// node_index ancestor(node_index i, level Delta) const;
// Requirement: Delta <= depth(i).
// Effect: RETURN floors{ (i + 1) / 2^Delta - 1 };

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::ancestor(
navigation_pile<T, Container, Compare>::node_index i, 
navigation_pile<T, Container, Compare>::level delta) const {
  return ((i + 1) >> delta) - 1;
  //return (i + 1) / pow_two(delta) - 1;
}


// node_index root() const;
// Effect: 
// level delta <- eta - ceils{ log_2 n };
// RETURN 2^delta - 1;

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::root() const {

  level delta;
  if (_n <= 1) {
    delta = _eta - 1;
  }
  else {
    delta = _eta - ceiling_log(_n);
  }

  return pow_two(delta) - 1;
}


// leaf_index first_leaf() const;
// Effect: RETURN 2^eta -  1;
/*
template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::leaf_index 
navigation_pile<T, Container, Compare>::first_leaf() const{
  return pow_two(_eta) - 1;
}
*/

// leaf_index last_leaf() const;
// Effect: RETURN 2^eta + n - 2;

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::leaf_index 
navigation_pile<T, Container, Compare>::last_leaf() const{
  //return pow_two(_eta) + _n - 2;
  return _first_leaf + _n - 1;
}

// New

template <class T, class Container, class Compare>
bool
navigation_pile<T, Container, Compare>::empty() const {
 return (m == 0);
}


// element_index size() const;
// Effect: RETURN n;

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::size_type
navigation_pile<T, Container, Compare>::size() const {
  return _n;
}


// bool is_first_child(node_index i) const;
// Effect: RETURN (i BITAND 1 = 1);

template <class T, class Container, class Compare>
bool 
navigation_pile<T, Container, Compare>::is_first_child(
navigation_pile<T, Container, Compare>::node_index i
) const {
  return (i & 1 == 1);
}


// bool is_root(node_index i) const;
// Effect: RETURN (i = root());

template <class T, class Container, class Compare>
bool 
navigation_pile<T, Container, Compare>::is_root(
navigation_pile<T, Container, Compare>::node_index i
) const {
  return i == _root;
}


// bool is_in_use(node_index i) const;
// Effect: 
// level gamma <- height(i);
// leaf_index start_of_leaf_sequence <- 2^gamma * i + 2^gamma - 1;
// RETURN (gamma <= ceils{ log_2 n } LOGICALAND start_of_leaf_sequence <= last_leaf());
/*
template <class T, class Container, class Compare>
bool 
navigation_pile<T, Container, Compare>::is_in_use(
navigation_pile<T, Container, Compare>::node_index i
) const {
  level gamma = height(i);
  leaf_index start_of_leaf_sequence = pow_two(gamma) * i + pow_two(gamma) - 1;
  return (gamma <= size_type(ilogb(_n) + 1) && start_of_leaf_sequence <= last_leaf());
}
*/

// bit_index start_of_offset(branch_index i) const;
// Effect: 
// level delta <- depth(i);
// level gamma <- height(i);
// branch_index index_at_own_level <- i - 2^delta + 1;
// bit_index bits_upto_this_level <- (eta + 1) * delta * (delta + 1) / 2 - delta * (delta + 1) * (2 * delta + 1) / 6;
// bit_index bits_before <- bits_upto_this_level  + index_at_own_level * gamma;
// RETURN bits_before;

// Not implemented


// leaf_index element_to_leaf(element_index ell) const;
// Effect: RETURN first_leaf() + ell;

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::leaf_index 
navigation_pile<T, Container, Compare>::element_to_leaf(
navigation_pile<T, Container, Compare>::element_position ell
) const {
  return _first_leaf + ell - &A[0];
}


// element_index leaf_to_element(leaf_index L) const;
// Effect: RETURN L - first_leaf();

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::element_position 
navigation_pile<T, Container, Compare>::leaf_to_element(
navigation_pile<T, Container, Compare>::leaf_index L
) const {
  return element_position(&A[0]) + L - _first_leaf;
}


// (element_index, offset) jump_to_element(branch_index i) const;
// Effect: 
// bit_index s <- start_of_offset(i);
// level gamma <- height(i);
// offset lambda <- B[s.. s + gamma);
// leaf_index start_of_leaf_sequence <- 2^gamma * i + 2^gamma -  1;
// element_index ell <- leaf_to_element(start_of_leaf_sequence + lambda);
// RETURN (ell, lambda);

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::element_position
navigation_pile<T, Container, Compare>::jump_to_element(
navigation_pile<T, Container, Compare>::branch_index i
) const {
  return B[i];
}


// navigation_pile(const input_sequence& X, element_index capacity,
// const ordering& f, const container_sequence& Y);
// Effect: 
// Allocate space for the sequence A[0.. capacity) of the
// same type as the sequence referred to by Y;
// Copy the elements stored in the sequence referred to by X to A;
// Construct a private copy of the ordering function referred to by f and call it less();
// Let 2^eta be the smallest power of 2 larger than or equal to capacity;
// Allocate space for the bit sequence B[0.. 2^eta+1);
// Use n as a synonym for A.size();
// IF (n <= 1) RETURN;
// make_pile();

template <class T, class Container, class Compare>
navigation_pile<T, Container, Compare>::navigation_pile(size_type N, 
							const Compare& x, 
							const Container& Y
							) : A(Y), B(int(pow_two(ilogb_zero(N - 1) + 1))), _n(Y.size()), _N(N), _eta(ilogb_zero(N - 1) + 1), _first_leaf(pow_two(_eta) - 1), _root(root()), _root_n(pow_two(height(_root) - 1)) {
  A.reserve(N);
  less = x;
  if (_n < 1) return;
  if (_n == 1) {
    B[_root] = &A[0];
    return;
  }
  make_pile();
}


template <class T, class Container, class Compare>
template <class InputIterator>
navigation_pile<T, Container, Compare>:: navigation_pile(InputIterator first, 
							 InputIterator last, 
							 size_type N,
							 const Compare& x,
							 const Container& Y
							 ) : A(Y), B(int(pow_two(ilogb_zero(N - 1) + 1))), _n((last - first) + Y.size()), _N(N),  _eta(ilogb_zero(N - 1) + 1), _first_leaf(pow_two(_eta) - 1), _root(root()), _root_n(pow_two(height(_root) - 1)) {
  A.reserve(N);
  A.insert(A.end(), first, last);
  less = x;
  if (_n < 1) return;
  if (_n == 1) {
    B[_root] = &A[0];
    return;
  }
  make_pile();
}


template <class T, class Container, class Compare>
navigation_pile<T, Container, Compare>::~navigation_pile(){
}


// The make_pile() function and its utility functions are as follows.

// void make_pile();
// Effect: 
// branch_index j <- parent(first_leaf());
// branch_index ell <- parent(last_leaf());
// FOR (branch_index k <- ell; k >= j; --k ) 
//   handle_height_one_branch(k);
//   branch_index i <- k;
//   WHILE (is_first_child(i) LOGICALAND NOT is_root(i)) 
//     i <- parent(i);
//     handle_upper_branch(i);

template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::make_pile() {
  branch_index i;
  branch_index j;
  branch_index k;
  element_position Pj;
  element_position Pk;
  element ej;
  element ek;
  node_index last_node_in_use;
  branch_index f = parent(_first_leaf);
  for (int e = int(parent(last_leaf())); e >= int(f); --e ) {
    handle_height_one_branch(e);
    j = e;
    Pj = B[j];
    ej = *Pj;    
    last_node_in_use = last_leaf();
    while (is_first_child(j) && !is_root(j)) {
      i = parent(j);
      last_node_in_use = parent(last_node_in_use);
      k = sibling(j);
      k = k - (k > last_node_in_use);
      Pk = B[k];
      ek = *Pk;
      if (less(ej, ek)) {
	B[i] = Pk;
	j = i;
	Pj = Pk;
	ej = ek;
      }
      else {
	B[i] = Pj;
	j = i;
      }
    }
  }
}


// void handle_height_one_branch(branch_index i);
// Effect: 
// bit_index s <- start_of_offset(i);
// element_index ell <- leaf_to_element(first_child(i));
// IF (is_in_use(second_child(i))) 
//   element_index m <- leaf_to_element(second_child(i));
//   B[s.. s + 1), <- less(A[m],A[ell]) ? 0 : 1;
// ELSE 
//   B[s.. s + 1), <- 0;

template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::handle_height_one_branch(
navigation_pile<T, Container, Compare>::branch_index i
) {
  node_index fc = first_child(i);
  node_index sc = fc + 1;
  element_position ell = leaf_to_element(fc);
  if (sc <= last_leaf()) {
    element_position m = leaf_to_element(sc);
    B[i] = less(*m, *ell) ? ell : m;
  }
  else {
    B[i] = ell;
  }
}


// void handle_upper_branch(branch_index i);
// Effect: 
// bit_index s <- start_of_offset(i);
// level gamma <- height(i);
// (element_index ell, offset lambda), <- jump_to_element(first_child(i));
// IF (is_in_use(second_child(i))) 
//   (element_index m, offset mu) <- jump_to_element(second_child(i));
//   B[s.. s + gamma), <- less(A[m],A[ell]) ? lambda : 2^(gamma-1) + mu;
// ELSE 
//   B[s.. s + gamma), <- lambda;
/*
template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::handle_upper_branch(
navigation_pile<T, Container, Compare>::branch_index i
) {
  element_position ell = jump_to_element(first_child(i));
  if (is_in_use(second_child(i))) {
    element_position m = jump_to_element(second_child(i));
    B[i] = less(*m, *ell) ? ell : m;
  }
  else { 
    B[i] = ell;
  }
}
*/

// The top() function is easily realized by following the
// offset stored at the root. Clearly, only O(1) instructions are needed.
// const element& top() const;
// Effect: 
// (element_index ell, cdot), <- jump_to_element(_root);
// RETURN A[ell];

template <class T, class Container, class Compare>
const typename navigation_pile<T, Container, Compare>::value_type& 
navigation_pile<T, Container, Compare>::top() const {
  element_position top_elem = jump_to_element(_root);
  return *top_elem;
}


// void push(const element& x);
// Effect: 
// A.push_back(x);
// node_index i <- binary_search_on_special_path(x);
// update_partial_path(last_leaf(), i, last_leaf());

template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::push(
const navigation_pile<T, Container, Compare>::value_type& x
) {
  if(_n >= _N) throw Overflow();
  level old_root = _root;
  insert_new_element(x);
  if(_n == 1) {
    B[_root] = &A[0];
    return;
  };
  if(_n - 1 == _root_n << 1) {
    _root_n = _root_n << 1;
    _root = parent(_root);
    B[_root] = B[old_root];
  }
  node_index j = last_leaf();
  if ((j & 1)) {
    j = update_partial_path(j);
  }
  node_index i = binary_search_on_special_path(j, x);
  
  // Alternativ to binary search
  //node_index i = bottom_up_on_special_path(j, x);
  
  update_partial_path(j, i, last_leaf());
}


// node_index binary_search_on_special_path(const element& x);
// Effect: 
// level Delta <- height(root());
// node_index j <- last_leaf();
// WHILE (Delta > 0)
//   level half <- floors{Delta / 2};
//   branch_index i <- ancestor(j, Delta - half);
//   (element_index ell, cdot), <- jump_to_element(i);
//   IF (less(A[ell], x))
//     j <- i;
//     Delta <- half;
//  ELSE
//    Delta <- Delta - half -  1;
// RETURN j;

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::binary_search_on_special_path(
navigation_pile<T, Container, Compare>::node_index j, 
const navigation_pile<T, Container, Compare>::element& x
) {
  level delta = depth(j) - depth(_root);
  branch_index i;
  level half;
  element_position ell;
  while (delta > 0) {
    half = delta / 2;
    i = ancestor(j, delta - half);
    ell = jump_to_element(i);
    if (less(*ell, x)) {
      j = i;
      delta = half;
    }
    else {
      delta = delta - half -  1;
    }
  }
  return j;
}


// void  update_partial_path(node_index j, node_index i, leaf_index K);
// Requirement: j is an ancestor of K and i is an ancestor of j.
// Effect: 
// level gamma <- height(j);
// WHILE (j != i)
//   j <- parent(j);
//   gamma <- gamma + 1;
//   bit_index s <- start_of_offset(j);
//   leaf_index start_of_leaf_sequence <- 2^gamma * j + 2^gamma - 1;
//   offset lambda <- K - start_of_leaf_sequence;
//   B[s.. s + gamma), <- lambda;

template <class T, class Container, class Compare>
void  
navigation_pile<T, Container, Compare>::update_partial_path(
navigation_pile<T, Container, Compare>::node_index j, 
navigation_pile<T, Container, Compare>::node_index i, 
navigation_pile<T, Container, Compare>::leaf_index K
) {
  element_position lambda ;
  while (j != i) {
    j = parent(j);
    lambda = K - _first_leaf + &A[0];
    B[j] = lambda;
  }
}


// ### NEW
template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::update_partial_path(
navigation_pile<T, Container, Compare>::leaf_index K
) {
  node_index j = K;
  node_index i = parent(j);
  node_index rootnode = _root;
  element_position lambda = last_leaf() - _first_leaf + &A[0];
  size_type le = _n - 1;
  while (i != rootnode && ((le & 1) == 0)) {
    le = le >> 1;
    B[i] = lambda;
    j = i;
    i = parent(i);
  }
  return j;
}

// void pop();
// Effect: 
// IF (n = 1)
//   A.pop_back();
//   RETURN;
// (element_index m, cdot) <- jump_to_element(root());
// const element_index ell <- leaf_to_element(last_leaf());
// (branch_index i_1, branch_index i_2) <- first_two_ancestors_with_more_than_one_child(last_leaf());
// (element_index k, cdot) <- jump_to_element(i_2);
// element_index j_1;
// IF (height(i_1) = 1)
//   j_1 <- ell-1;
// ELSE
//   (j_1, cdot) <- jump_to_element(first_child(i_1));
// IF (m = ell)
//   A.pop_back();
//   update_full_path(last_leaf());
// ELSEIF (k != ell) 
//   A[m] <- A[ell];
//   A.pop_back();
//   update_partial_path(i_1, i_2, element_to_leaf(j_1), last_leaf());
//   update_full_path(element_to_leaf(m)); 
// ELSE 
//   (element_index j_2, cdot) <- jump_to_element(first_child(i_2));
//   A[m] <- A[j_2];
//   A[j_2] <- A[ell];
//   A.pop_back();
//   update_partial_path(i_1, i_2, element_to_leaf(j_1), last_leaf());
//   update_partial_path(i_2, root(), element_to_leaf(j_2), last_leaf());
//   update_full_path(element_to_leaf(m));

/*
// Pop using second ansector technique
template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::pop() {
  if(_n == 0) throw Underflow();
  if (_n == 1) {
    erase_last_element();
    return;
  }
  leaf_index ll = last_leaf();
  element_position m = jump_to_element(_root);
  const element_position ell = leaf_to_element(ll);
  branch_index i_1;
  branch_index i_2;
  first_two_ancestors_with_more_than_one_child(i_1, i_2, ll);
  element_position k = jump_to_element(i_2);
  element_position j_1;
  if (height(i_1) == 1) {
    j_1 = ell - 1;
  }
  else {
    j_1 = jump_to_element(first_child(i_1));
  }
  bool new_root = (_n - 1 == _root_n && _n - 1 != 1);
  if (m == ell) {
    erase_last_element();
    if (new_root) {
      _root_n = _root_n >> 1;
      _root = first_child(_root);
    }
    update_full_path(last_leaf());
  }
  else {
    if (k != ell) { 
      *m = *ell;
      erase_last_element();
      if (new_root) {
	_root_n = _root_n >> 1;
	_root = first_child(_root);
      }
      update_partial_path(i_1, i_2, element_to_leaf(j_1), ll);
      update_full_path(element_to_leaf(m)); 
    }
    else { 
      element_position j_2 = jump_to_element(first_child(i_2));
      *m = *j_2;
      *j_2 = *ell;
      erase_last_element();
      if (new_root) {
	_root_n = _root_n >> 1;
	_root = first_child(_root);
      }
      update_partial_path(i_1, i_2, element_to_leaf(j_1), ll);
      update_partial_path(i_2, _root, element_to_leaf(j_2), ll);
      update_full_path(element_to_leaf(m));
    }
  }
}
*/


// Pop using first ansector technique
template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::pop() {
  if(_n == 0) throw Underflow();
  if (_n == 1) {
    erase_last_element();
    return;
  }
  leaf_index ll = last_leaf();
  element_position k;
  element_position m = jump_to_element(_root);
  const element_position ell = leaf_to_element(ll);
  branch_index i;
  i = first_ancestor_with_two_children(ll);
  
  if (height(i) == 1) {
    k = ell - 1;
  }
  else {
    k = jump_to_element(first_child(i));
  }
  bool new_root = (_n - 1 == _root_n && _n - 1 != 1);
  
  if (ell == m) {
    erase_last_element();
    if (new_root) {
      _root_n = _root_n >> 1;
      _root = first_child(_root);
    }
    update_full_path(last_leaf());
  }
  else if (jump_to_element(i) == ell) {
    *m = *k;
    *k = *ell;
    erase_last_element();
    if (new_root) {
      _root_n = _root_n >> 1;
      _root = first_child(_root);
    }
    leaf_index lk = element_to_leaf(k);
    if (B[parent(lk)] == ll + &A[0]) B[parent(lk)] = lk + &A[0];
    update_partial_path(i, _root, lk, ll);
    update_full_path(element_to_leaf(m));
  }
  else {
    *m = *ell;
    erase_last_element();
    if (new_root) {
      _root_n = _root_n >> 1;
      _root = first_child(_root);
    }
    update_full_path(element_to_leaf(m));
  }
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::first_ancestor_with_two_children(
navigation_pile<T, Container, Compare>::leaf_index L
) {
  size_type le = _n - 1;
  unsigned int i = 1;

  while ((le & 1) == 0) {
    i++;
    le = le >> 1;
  }
  return (ancestor(L, i));
}


// (branch_index, branch_index) first_two_ancestors_with_more_than_one_child(leaf_index L) const;
// Requirement: n > 1.
// Effect: 
// branch_index i_1 <- parent(L);
// WHILE  (NOT is_in_use(second_child(i_1)))
//   i_1 <- parent(i_1);
// IF (is_root(i_1))
//   RETURN (i_1, i_1);
// branch_index i_2 <- parent(i_1);
// WHILE  (NOT is_in_use(second_child(i_2)))
//   i_2 <- parent(i_2);
// RETURN (i_1, i_2);

template <class T, class Container, class Compare>
void navigation_pile<T, Container, Compare>::first_two_ancestors_with_more_than_one_child(
  navigation_pile<T, Container, Compare>::branch_index &i_1, 
  navigation_pile<T, Container, Compare>::branch_index &i_2,
  navigation_pile<T, Container, Compare>::leaf_index L
) {
  size_type le = _n - 1;
  unsigned int i = 1;
  while ((le & 1) == 0) {
    i++;
    le = le >> 1;
  }
  i_1 = ancestor(L, i);
  if (is_root(i_1)) {
    i_2 = i_1;
    return;
  }
  i = i + 1;
  le = le >> 1;
  while ((le & 1) == 0) {
    i++;
    le = le >> 1;
  }
  i_2 = ancestor(L, i);
}


// void update_full_path(leaf_index L);
// Requirement: n > 1.
// Effect: 
// branch_index i <- parent(L);
// handle_height_one_branch(i);
// WHILE (NOT is_root(i)) 
//   i <- parent(i);
//   handle_upper_branch(i);

template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::update_full_path(
navigation_pile<T, Container, Compare>::leaf_index L
) {
  branch_index i = parent(L);
  handle_height_one_branch(i);
  node_index last_node_in_use = last_leaf();
  node_index fc;
  node_index sc;

  while (!is_root(i)) {
    i = parent(i);
    fc = first_child(i);
    sc = fc + 1;
    last_node_in_use = parent(last_node_in_use);
    element_position ell = jump_to_element(fc);
    if (sc <= last_node_in_use) {
      element_position m = jump_to_element(sc);
      B[i] = less(*m, *ell) ? ell : m;
    }
    else { 
      B[i] = ell;
    }
  }

  /*
  branch_index i;
  branch_index k;
  element_position Pj;
  element_position Pk;
  element ej;
  element ek;
  branch_index j = parent(L);
  handle_height_one_branch(j);
  node_index last_node_in_use = last_leaf();
  Pj = B[j];
  ej = *Pj;
  while (!is_root(j)) {
    i = parent(j);
    last_node_in_use = parent(last_node_in_use);
    k = sibling(j);
    k = k - (k > last_node_in_use);
    Pk = B[k];
    ek = *Pk;
    if (less(ej, ek)) {
      B[i] = Pk;
      j = i;
      Pj = Pk;
      ej = ek;
    }
    else {
     B[i] = Pj;
     j = i;
    }
  }
  */
}


// void update_partial_path(branch_index j, branch_index i, leaf_index K, leaf_index L);
// Requirement: i is an ancestor of j, K and L are in the leaf sequence of j.
// Effect: 
// WHILE (i != j)
//   bit_index s <- start_of_offset(j);
//   level gamma <- height(j);
//   offset kappa <- B[s.. s + gamma);
//   leaf_index start_of_leaf_sequence <- 2^gamma * j + 2^gamma - 1;
//   offset lambda <- L - start_of_leaf_sequence;
//   IF (kappa != lambda) RETURN;
//   offset mu <- K - start_of_leaf_sequence;
//   B[s.. s + gamma), <- mu;
//   j <- parent(j);

template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::update_partial_path(
navigation_pile<T, Container, Compare>::branch_index j,
navigation_pile<T, Container, Compare>::branch_index i, 
navigation_pile<T, Container, Compare>::leaf_index K, 
navigation_pile<T, Container, Compare>::leaf_index L
) {
  element_position mu = K - _first_leaf + &A[0];
  element_position kappa;
  element_position lambda = L - _first_leaf + &A[0];
  while (i != j) {
    kappa = B[j];
    if (kappa != lambda) return;
    B[j] = mu;
    j = parent(j);
  }
}


template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::insert_new_element(
const navigation_pile<T, Container, Compare>::element& z
) {
  A.push_back(z);
  _n++;
}


template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::erase_last_element() {
  A.pop_back();
  _n--;
}

// Calculates Sibling

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index
navigation_pile<T, Container, Compare>::sibling(
navigation_pile<T, Container, Compare>::node_index i
) const {
  bool odd = i & 1;
  return (i + odd - !odd);
}


// Calculates power z of two

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::pow_two(
navigation_pile<T, Container, Compare>::level z
) const {
  return (1 << z);
  //return int(pow(2, z));
}


// Calculates ceils{ log_2 n }

template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::size_type 
navigation_pile<T, Container, Compare>::ceiling_log(
navigation_pile<T, Container, Compare>::size_type n
) const {
  size_type i = ilogb(n);
  return i + (pow_two(i) != n);
}


// Calculates ilogb but returning zero if i <= 0

template <class T, class Container, class Compare>
int
navigation_pile<T, Container, Compare>::ilogb_zero(
double i
) const {
  if (i <= 0) return 0;
  return ilogb(i);
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::node_index 
navigation_pile<T, Container, Compare>::bottom_up_on_special_path(
navigation_pile<T, Container, Compare>::node_index j, 
const navigation_pile<T, Container, Compare>::element& x
) {
  element_position ell;
  branch_index i = parent(j);
  while (i >= _root) {
    ell = jump_to_element(i);
    if (less(*ell, x)) {
      j = i;
      if (i == 0) return j;  // parent(0) != -1
      i = parent(i);
    }
    else {
      return j;
    }
  }
  return j;   
}


#include <is_sorted.c++> // defines cphstl::is_sorted 

// Utility function used for testing

template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::sort_pile(
size_type n, 
const Compare& x
) {
  less = x;
  Container t(n);
  std::copy(A.begin(), A.end(), t.begin());
  std::sort(t.begin(), t.end(), less);
  Container E(n);
  // use i-- insted of int i = 0 and n++ to make is_sorted work
  int i = n - 1;  
  while(size() > 0) {
    E[i] = top();
    pop();
    i--;
  }

  assert(cphstl::is_sorted(E.begin(), E.end(), less));
  assert(std::equal(E.begin(), E.end(), t.begin()));
}

