#ifndef __HEAP_FUNCTIONS__
#define __HEAP_FUNCTIONS__

#ifndef __HEAP_POLICY__
#include "heap_policy.cpp" // defines the heap policy
#endif
#ifndef __UTILITY_FUNCTIONS__
#include "utility_functions.cpp" // defines the functions needed by heap functions
#endif

template <int d, typename position, typename ordering, typename policy>
void pop_heap(position a, const ordering& less, policy& p) {
  typedef typename policy::index index;
  typedef typename policy::element element;
  index i = p.root();
  index j = p.last_leaf();
  element x = *(a + j);
  *(a + j) = *(a + i);
  p.erase_last_leaf();
  sift_down<d, position, index, element, ordering, policy>(a, i, x, less, p);
}

template <int d, typename position, typename ordering, typename policy>
void push_heap(position a, const ordering& less, policy& p) {
  typedef typename policy::index index;
  typedef typename policy::element element;
  p.insert_new_leaf();
  index j = p.last_leaf();
  sift_up<d, position, index, element, ordering, policy>(a, j, a[j], less, p);
}

template <int d, typename position, typename ordering, typename policy>
void make_heap(position a, const ordering& less, policy& p) {
  if (p.size() < 2) return;
  //std::cout << "before make_heap: ";
  //show(a, a + p.size());
  typedef typename policy::index index;
  typedef typename policy::element element;
  const index m = p.leftmost_node(p.depth(p.last_leaf()));
  const index j = p.parent(m);
  const index k = p.parent(p.last_leaf());
  
  //std::cout << "inside j, k, m: " << j << " " << k << " " << m << std::endl;

  for (index l = m - 1; l >= j; --l) {
    if (l <= k) {
      sift_down<d, position, index, element, ordering, policy>(a, l, *(a + l), less, p);
    }
    for (index i = l; p.is_first_child(i); ) {
      i = p.parent(i);
      sift_down<d, position, index, element, ordering, policy>(a, i, *(a + i), less, p);
    }
  }
  //std::cout << "after make_heap: ";
  //show(a, a + p.size());
}

template <int d, typename position, typename ordering, typename policy>
void sort_heap(position a, const ordering& less, policy& p) {
  while (p.size() > 1) {
    pop_heap<d, position, ordering, policy>(a, less, p);
  }
}

#include <is_sorted.c++> // defines cphstl::is_sorted
#include "debugging.c++" // defines show

template <int d, typename position, typename ordering>
bool is_heap(position a, position b, const ordering& less) {
  heap_policy<d, position, ordering> p(b - a);
  if (p.size() < 2) return true;
  typedef typename heap_policy<d, position, ordering>::index index;

  index last_branch = p.parent(p.last_leaf());
  for (index i = p.root(); i <= last_branch; ++i) {
    index j = p.first_child(i);
    index k = j + index(d) - 1;
    k = (k <= p.last_leaf()) ? k : p.last_leaf();
    while (j <= k) {
      if (less(*(a + i), *(a + j))) return false;
      ++j;
    }
  }
  return true;
}

#include <algorithm>
#include <vector>

template <int d, typename position, typename ordering>
void heapsort(position a, position b, const ordering& less) {
  typedef heap_policy<d, position, ordering> policy;
  policy p(b - a);

  typedef typename policy::element element;
  std::vector<element> t(b - a);
  std::copy(a, b, t.begin());
  std::sort(t.begin(), t.end(), ordering());

  //std::cout << "before make_heap: ";
  //show(a, b);

  make_heap<d, position, ordering, policy>(a, less, p);
  bool result = is_heap<d, position, ordering>(a, b, less);
  assert(result);

  //std::cout << "before sort_heap: ";
  //show(a, b);

  sort_heap<d, position, ordering, policy>(a, less, p);

  //std::cout << "after sort_heap: ";
  //show(a, b);

  assert(cphstl::is_sorted(a, b, less));
  assert(std::equal(a, b, t.begin()));
}

template <int d, typename position, typename ordering, typename policy>
void sort(position a, position b, const ordering& less) {
  policy p(b - a);

  make_heap<d, position, ordering, policy>(a, less, p);
  sort_heap<d, position, ordering, policy>(a, less, p);
}

#endif // __HEAP_FUNCTIONS__
