#ifndef __HEAP_POLICY__
#define __HEAP_POLICY__

#include "heap_policy.hpp" // declares heap_policy
#include <cmath> // defines std::ilogb

extern int ilogb(double) throw();

#ifdef BUFFER
template <int d, typename position, typename ordering>
heap_policy<d, position, ordering>::heap_policy (
  heap_policy<d, position, ordering>::index n
): n(n) {
  if (n <= index(d) + 1) { l = 1;}
  else {
    l = d * ((n - 1) / index(d) - 1) + 1;
    if ( l + index(d) != n) {
      l = l + index(d);
    }
  }
}
#else
template <int d, typename position, typename ordering>
heap_policy<d, position, ordering>::heap_policy (
  heap_policy<d, position, ordering>::index n
): n(n) {
}
#endif

template <int d, typename position, typename ordering>
bool
heap_policy<d, position, ordering>::is_root (
  heap_policy<d, position, ordering>::index i
) const {
  return (i == 0);
}

template <int d, typename position, typename ordering>
bool
heap_policy<d, position, ordering>::is_first_child (
  heap_policy<d, position, ordering>::index i
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  return (i - (i / index(d)) * index(d) == 1);
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::size () const {
  return n;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::level
heap_policy<d, position, ordering>::depth (
  heap_policy<d, position, ordering>::index i
) const {
  typedef typename heap_policy<d, position, ordering>::level level;
  typedef typename heap_policy<d, position, ordering>::index index;
  level h = 0;
  index m = 1;
  index nr = 0;
  while (i > nr) {
    m = m * index(d);
    nr = nr + m;
    ++h;
  }
  return h;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::root () const {
  return 0;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::leftmost_node (
  heap_policy<d, position, ordering>::level h
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  index r = 1;
  while (h > 0) {
    r = r * index(d);
    --h;
  }
  r = (r - 1) / index(d - 1);
  return r;
}

#ifdef BUFFER
template <int d, typename position, typename ordering>
heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::last_leaf () const {
  return l - 1;
}
#else
template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::last_leaf () const {
  return n - 1;
}
#endif

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::first_child (
  heap_policy<d, position, ordering>::index p
) const {
  return d * p + 1;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::parent (
  heap_policy<d, position, ordering>::index c
) const {
  return (c - 1) / d;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::ancestor (
  heap_policy<d, position, ordering>::index j,
  heap_policy<d, position, ordering>::level h,
  heap_policy<d, position, ordering>::level Delta
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  index t = 1;
  while (Delta > 0) {
    t = t * index(d);
    --Delta;
  }
  index i = (t - 1) / index(d - 1);
  return (j - i) / t;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::top_some_absent (
  position a,
  heap_policy<d, position, ordering>::index i,
  const ordering& less
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  typedef typename heap_policy<d, position, ordering>::element element;
  index j = heap_policy<d, position, ordering>::first_child(i);
  index k = (j + index(d) < n) ? j + index(d) : n;
  index maxc = j;
  element max = a[j];
  ++j;
  while (j < k) {
    if (less(max, a[j])) {
      maxc = j;
      max = a[j];
    }
    ++j;
  }
  return maxc;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::top_all_present (
  position a,
  heap_policy<d, position, ordering>::index i,
  const ordering& less
) const {
  return heap_policy<d, position, ordering>::top_some_absent(a, i, less);
}

template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::update (
  position a,
  heap_policy<d, position, ordering>::index i,
  const heap_policy<d, position, ordering>::element& x
) {
  a[i] = x;
}

#ifdef BUFFER
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::erase_last_leaf (
) {
  --n;
  if (l == n) {
    if ( n <= index(d) ) {l = 1;}
    else {l = l - index(d);}
  }
}
#else
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::erase_last_leaf (
) {
  --n;
}
#endif

#ifdef BUFFER
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::insert_new_leaf (
) {
  ++n;
  if (( l + 2 * index(d)) == n ) {
    l = l + index(d);
  }
}
#else
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::insert_new_leaf (
) {
  ++n;
}
#endif

/* 

Specialization for d = 2.

*/

#ifdef BUFFER
template <> template <typename position, typename ordering>
heap_policy<2, position, ordering>::heap_policy (
  heap_policy<2, position, ordering>::index n
): n(n){
  if (n <= 3 ) { l = 1;}
  else {
    l = ((((n - 1) >> 1) - 1) << 1) + 1;
    if ( l + 2 != n) {
      l = l + 2;
    }
  }
}
#else
template <> template <typename position, typename ordering>
heap_policy<2, position, ordering>::heap_policy (
  heap_policy<2, position, ordering>::index n
): n(n) {
}
#endif

template <> template <typename position, typename ordering>
bool
heap_policy<2, position, ordering>::is_root (
  heap_policy<2, position, ordering>::index i
) const {
  return (i == 0);
}

template <> template <typename position, typename ordering>
bool
heap_policy<2, position, ordering>::is_first_child (
  heap_policy<2, position, ordering>::index i
) const {
  typedef typename heap_policy<2, position, ordering>::index index;
  return (i & 1) == 1;
}

template <> template <typename position, typename ordering>
typename heap_policy<2, position, ordering>::index
heap_policy<2, position, ordering>::size () const {
  return n;
}

template <> template <typename position, typename ordering>
typename heap_policy<2, position, ordering>::level
heap_policy<2, position, ordering>::depth (
  heap_policy<2, position, ordering>::index i
) const {
  typedef typename heap_policy<2, position, ordering>::level level;
  return level(ilogb(1 + i));

  /*
  //Alternativ 1
  const static index index_level[33] = {
    0, 1, 3, 7, 15, 31, 63, 127, 255, 511, 1023, 2047, 4095, 8191, 16383, 32767, 65535, 131071, 262143, 524287, 1048575, 2097151, 4194303, 8388607, 16777215, 33554431, 67108863, 134217727, 268435455, 536870911, 1073741823, 2147483647, INT_MAX
  }; // 0 and INT_MAX added as padding

  level s = 0;

  if (index_level[16] <= i)  s = s + 16;
  if (index_level[s + 8] <= i)  s = s + 8;
  if (index_level[s + 4] <= i)  s = s + 4;
  if (index_level[s + 2] <= i)  s = s + 2;
  if (index_level[s + 1] <= i)  s = s + 1;

  return(s);
  */

  /*
  //Alternativ 2

  if (i < 1) return level(0);
  if (i < 3) return level(1);
  if (i < 7) return level(2);
  if (i < 15) return level(3);
  if (i < 31) return level(4);
  if (i < 63) return level(5);
  if (i < 127) return level(6);
  if (i < 255) return level(7);
  if (i < 511) return level(8);
  if (i < 1023) return level(9);
  if (i < 2047) return level(10);
  if (i < 4095) return level(11);
  if (i < 8191) return level(12);
  if (i < 16383) return level(13);
  if (i < 32767) return level(14);
  if (i < 65535) return level(15);
  if (i < 131071) return level(16);
  if (i < 262143) return level(17);
  if (i < 524287) return level(18);
  if (i < 1048575) return level(19);
  if (i < 2097151) return level(20);
  if (i < 4194303) return level(21);
  if (i < 8388607) return level(22);
  if (i < 16777215) return level(23);
  if (i < 33554431) return level(24);
  if (i < 67108863) return level(25);
  if (i < 134217727) return level(26);
  if (i < 268435455) return level(27);
  if (i < 536870911) return level(28);
  if (i < 1073741823) return level(29);
  if (i < 2147483647) return level(20);  
  else return level(31);
  */
}

template <> template <typename position, typename ordering>
typename heap_policy<2, position, ordering>::index
heap_policy<2, position, ordering>::root () const {
  return 0;
}

template <> template <typename position, typename ordering>
typename heap_policy<2, position, ordering>::index
heap_policy<2, position, ordering>::leftmost_node (
  heap_policy<2, position, ordering>::level h
) const {
  typedef typename heap_policy<2, position, ordering>::index index;
  const static index index_level[33] = {
    0, 1, 3, 7, 15, 31, 63, 127, 255, 511, 1023, 2047, 4095, 8191, 16383, 32767, 65535, 131071, 262143, 524287, 1048575, 2097151, 4194303, 8388607, 16777215, 33554431, 67108863, 134217727, 268435455, 536870911, 1073741823, 2147483647, INT_MAX
  };

  return (index_level[h]);
}

#ifdef BUFFER
template <> template <typename position, typename ordering>
heap_policy<2, position, ordering>::index
heap_policy<2, position, ordering>::last_leaf () const {
  return l - 1;
}
#else
template <> template <typename position, typename ordering>
typename heap_policy<2, position, ordering>::index
heap_policy<2, position, ordering>::last_leaf () const {
  return n - 1;
}
#endif

template <> template <typename position, typename ordering>
typename heap_policy<2, position, ordering>::index
heap_policy<2, position, ordering>::first_child (
  heap_policy<2, position, ordering>::index p
) const {
  return (p << 1) + 1;
}

template <> template <typename position, typename ordering>
typename heap_policy<2, position, ordering>::index
heap_policy<2, position, ordering>::parent (
  heap_policy<2, position, ordering>::index c
) const {
  return (c - 1) >> 1;
}

template <> template <typename position, typename ordering>
typename heap_policy<2, position, ordering>::index
heap_policy<2, position, ordering>::ancestor (
  heap_policy<2, position, ordering>::index i,
  heap_policy<2, position, ordering>::level h,
  heap_policy<2, position, ordering>::level Delta
) const {
  return ((i + 1) >> Delta) - 1;
}

template <> template <typename position, typename ordering>
typename heap_policy<2, position, ordering>::index
heap_policy<2, position, ordering>::top_some_absent (
  position a,
  heap_policy<2, position, ordering>::index i,
  const ordering& less
) const {
  typedef typename heap_policy<2, position, ordering>::index index;
  index maxc = heap_policy<2, position, ordering>::first_child(i);
  if (maxc + 1 <= heap_policy<2, position, ordering>::last_leaf()) {
    if (less(*(a + maxc), *(a + maxc + 1))) {
      maxc = maxc + 1;
    }
  }
  return maxc;  
}

template <> template <typename position, typename ordering>
typename heap_policy<2, position, ordering>::index
heap_policy<2, position, ordering>::top_all_present (
  position a,
  heap_policy<2, position, ordering>::index i,
  const ordering& less
) const {
  typedef typename heap_policy<2, position, ordering>::index index;
  index j = heap_policy<2, position, ordering>::first_child(i);
  
  if (less(a[j], a[j + 1])) j++;
  return j;
}

template <> template <typename position, typename ordering>
void
heap_policy<2, position, ordering>::update (
  position a,
  heap_policy<2, position, ordering>::index i,
  const heap_policy<2, position, ordering>::element& x
) {
  *(a + i) = x;
}

#ifdef BUFFER
template <> template <typename position, typename ordering>
void
heap_policy<2, position, ordering>::erase_last_leaf (
) {
  --n;
  if (l == n) {
    if (n <= 2 ) { l = 1;}
    else {l = l - 2;}
  }
}
#else
template <> template <typename position, typename ordering>
void
heap_policy<2, position, ordering>::erase_last_leaf (
) {
  --n;
}
#endif

#ifdef BUFFER
template <> template <typename position, typename ordering>
void
heap_policy<2, position, ordering>::insert_new_leaf (
) {
  ++n;
  if (( l + 4) == n ) {
    l = l + 2;
  }
}
#else
template <> template <typename position, typename ordering>
void
heap_policy<2, position, ordering>::insert_new_leaf (
) {
  ++n;
}
#endif

#endif // __HEAP_POLICY__
