#ifndef __measure_execution_time__
#define __measure_execution_time__

/*
These tools can be used to measure the execution time of a program

Author: Jyrki Katajainen
Email: jyrki@diku.dk
Version: $Revision: 1.4 $
Date: $Date: 2003/01/28 12:04:19 $

When writing this module I used the following articles as my inspiration

Andrew Koenig and Barbara E. Moo, Performance: Myths, Measurements, 
and Morals, Part 5: How Not to Measure Execution Time, Journal of 
Object-Oriented Programming 13,1 (2000), 29-31, and 39.

Andrew Koenig and Barbara E. Moo, Performance: Myths, Measurements, 
and Morals, Part 6 Useful Measurements---Finally!, Journal of 
Object-Oriented Programming 13,2 (2000), 31-33.

However, I had to extend their framework to handle side-effects.
*/

#include <ctime> // defines std::clock_t and std::clock()
#include <cstdlib> // defines std::exit()
#include <vector> // defines std::vector
#include <algorithm> //defines std::sort()
#include <iostream>  // defines standard streams

/* 

10 ms is a reasonable clock precision on most computers
1 s should be relatively large compared to one tick

*/

const std::clock_t reasonable_precision = CLOCKS_PER_SEC / 100;
const std::clock_t high_precision = 1 * CLOCKS_PER_SEC;

double clock_overhead() {
  std::clock_t k = std::clock();
  std::clock_t start, limit;

  // Wait for the clock to tick
  do  
    start = std::clock();
  while (start == k);

  // Interrogate the clock until it has advanced
  limit = start + reasonable_precision;

  unsigned long r = 0;
  while ((k = std::clock()) < limit)
    ++r;
  return double(k - start) / r;
}

/* 

If the execution time is more than high_precision, the experiment is performed
only once. Otherwise, the experiment is repeated several times until 90% of 
the  outcomes differ at most 20% from the median, or more than 100 trials
have been done after which the experiment is aborted.

*/

const unsigned int factor = 10;
const unsigned int percent = 20;
const unsigned int repetitions = 100;

/* 

I use a macro here because I want that the user can freely select the class 
name. Now only the constructor call needed to be given to the system.

*/

#define MEASURE(running_time, function) { \
  unsigned long clock_calls = 1; \
  double overhead = clock_overhead(); \
  std::clock_t start; \
  std::clock_t stop; \
  std::clock_t init = std::clock(); \
  do \
    start = std::clock(); \
  while (start == init); \
  function(); \
  std::clock_t t = std::clock(); \
  do \
    ++clock_calls; \
  while ((stop = std::clock()) == t); \
  double ticks = double(stop - start) - (double(clock_calls) * overhead); \
  if (ticks >= high_precision) { \
    running_time = ticks / double(CLOCKS_PER_SEC); \
  } \
  else { \
    unsigned int trial = 0; \
    std::vector<double> mv; \
    bool done = false; \
    do { \
      ++trial; \
      mv.reserve(mv.size() + 2 * factor); \
      std::clock_t k = std::clock(); \
      std::clock_t start; \
      do \
        start = std::clock(); \
      while (start == k); \
      for (unsigned int i = 2 * factor; i > 0; --i) { \
        unsigned long calls = 0; \
        unsigned long limit = 1; \
        unsigned long clock_calls = 0; \
        std::clock_t clocklimit = start + reasonable_precision; \
        std::clock_t t; \
        do { \
	  while (calls < limit) { \
	    function(); \
	    ++calls; \
	  } \
	  limit *= 2; \
	  ++clock_calls; \
        } while ((t = std::clock()) < clocklimit); \
        std::clock_t stop; \
        do \
	  ++clock_calls; \
        while ((stop = std::clock()) == t); \
        double ticks = \
          (double(stop - start) - (double(clock_calls) * overhead)) / double(calls); \
        mv.push_back(ticks / double(CLOCKS_PER_SEC)); \
        start = stop; \
      } \
      std::sort(mv.begin(), mv.end()); \
      double median = (mv[trial * factor] + mv[trial * factor - 1]) / 2; \
      if (mv[trial] > (median * (100.0 - percent)) / 100.0 && \
          mv[mv.size() - trial - 1] < (median * (100.0 + percent)) / 100.0) { \
        running_time = median; \
        done = true; \
        break; \
      }; \
    } while (trial < repetitions); \
    if (!done) { \
      std::cerr << "Less than " \
	        << 100 - factor \
	        << "% of the outcomes differ at most " \
	        << percent \
	        << "% from the median; try again" \
	        << std::endl; \
      std::exit(1); \
    }; \
  } \
}

template <typename functor>
double measure_primal(functor e) {
  double running_time = 0.0;
  MEASURE(running_time, e.primal)
  return running_time;
}

template <typename functor>
double measure_dual(functor e) {
  double running_time = 0.0;
  MEASURE(running_time, e.dual)
  return running_time;
}

/* Unittest for the timing routine


*/

#endif // __measure_execution_time__

