#!/usr/bin/env python2.2

"""
This Python module implements Jyrki's benchmarking framework for C++ programs

When creating this module, I used Python's unittest module as my inspiration.
For further details, see Python Library Reference.
This module has been extended by Christian Ulrik Sttrup & Jakob Pedersen.
The extensions include the posibility of running benchmarks remotely, 
measuring page fault and measuring cache misses. 
"""

__author__ = "Jyrki Katajainen, Christian Ulrik Sttrup & Jakob Pedersen"
__email__ = "jyrki@diku.dk, chrulle@diku.dk & bbrother@diku.dk"
__version__ = "$Revision: 1.27 $"[11:-2]
__date__ = "$Date: 2003/06/08 15:22:40 $"[7:-2]
__tool__ = "Benchmark"

import commands
import fnmatch
import os
import random
import string
import sys
import time
import traceback
import types
import copy

######################################################################
# result
######################################################################

class result:
  """
  An instance collecting the results of a benchmark run (case or suite)

  The output of a case is a tuple and that of a suite a list of tuples.
  All error messages are also saved.
  """

  def __init__(self):
    """
    initializes data structurs
    """
    self.answers = []
    self.errors = []
    self.should_abort = 0

  def __getinitargs__(self):
    return ()

  def start(self, benchmark):
    pass

  def stop(self, benchmark):
    pass

  def abort(self, benchmark):
    """
    Indicates that the whole benchmark run should be aborted
    """
    self.should_abort = 1

  def add_answer(self, benchmark, answer):
    """
    appends answer tuple to answer list
    """
    self.answers.append((benchmark, answer))

  def add_error(self, benchmark, error):
    """
    appends error tuple to error list
    """
    self.errors.append((benchmark, error))

  def run_cases(self):
    """
    counts the number of cases in a list of case/suite tuples
    """
    number_of_cases = 0
    for t in self.answers:
      number_of_cases += self.count_cases(t)
    return number_of_cases

  def count_cases(self, tuple):
    """
    counts the number of cases in a tuple recursively (a suite may contain a suite of suites and so on)
    """
    b, s = tuple
    if isinstance(b, case):
      return 1
    elif isinstance(b, suite):
      n = 0
      for i in s:
        n += self.count_cases(i)
      return n

  def number_of_errors(self):
    """
    counts errors in a list of error tuples
    """
    count = 0
    for t in self.errors:
      count += self.count_errors(t)
    return count

  def count_errors(self, tuple):
    """
    counts the number of errors in a tuple recursively
    """
    b, s = tuple
    if isinstance(b, case):
      return 1
    elif isinstance(b, suite) and isinstance(s, types.StringType):
      return 1
    elif isinstance(b, suite) and isinstance(s, types.ListType):
      n = 0
      for i in s:
        n += self.count_errors(i)
      return n

######################################################################
# case
######################################################################

class case:
  """
  A class whose instances are single benchmark cases
  """

  def __init__(self):
    """
    initializes datastructures. 
    sets randomized id (aproximation of unique id).
    sets default values for options.
    """
    self.driver_file = None
    self.counted_function = None
    self.driver_output = ''
    self.papitype = None
    self.constructor_call = None
    self.id = str(random.random())[2:] # must be unique
    self.title = ''
    self.working_directory = os.path.abspath(os.getcwd())
    self.include_files = []
    self.copy_include_files = []
    self.include_paths = \
      [self.working_directory, os.path.dirname(os.path.abspath(sys.argv[0]))]
    pythonpath = string.split(os.environ['PYTHONPATH'], ':') 
    for f in pythonpath:
      g = os.path.abspath(f)
      if os.path.isdir(g):
        self.include_paths.append(g)
        helpers = os.path.abspath(g + '/Helperbase')
        if os.path.exists(helpers):
          self.include_paths.append(helpers)
          list = os.listdir(helpers)
          for h in list:
            i = os.path.abspath(helpers + '/' + h)
            if os.path.isfile(i) and os.path.splitext(i)[1] == '.c++':
              self.include_files.append(i)
    self.remote_include_paths = []
    self.papi_path = None
    self.include_files.append('<iostream>')
    self.include_statements = None
    self.computer = os.uname()[1]
    self.gateway = None
    self.connection_protocol = "rsh"
    self.transfer_protocol = "rcp"
    self.compiler = "g++"
    self.compiler_options = ["-Wall", "-W", "-pedantic","-ansi"]
    self.compile_command = None
    self.execute_command = None
    self.dual_exists = None
    self.include_files_copied = 0
    self.remote_benchmark_dir ="./CPHSTL-Benchmarking-" + self.id + "/"
    self.remote_tidy_up = 1
    self.filenames_translated = 0
    self.time_unit = "s"
    self.time_factors = \
      {"ns": "1.0e+09", "micros": "1.0e+06", "ms": "1.0e+03", "s": "1.0", \
      "min": "(1.0/60.0)", "h": "(1.0/3600.0)", "day": "(1.0/86400.0)"}

  def copy_include_files_to_host(self, results, copied_files = []):
    """
    This function copies files needed to compile the benchmark on a remote computer to
    that computer through a gateway computer if specified.
    """
    status = 1
    if not self.connection_protocol:
      results.add_error(self, "No connection protocol given")
      return (status, copied_files)
    if not self.transfer_protocol:
      results.add_error(self, "No connection protocol or transfer protocol given")
      return (status, copied_files)
    shell_command = self.connection_protocol + " " + self.computer + " cd " + self.remote_benchmark_dir
    if self.gateway:
      shell_command = self.connection_protocol + " " + self.gateway + " " + shell_command
    (status, output) = commands.getstatusoutput(shell_command)
    if status != 0:
      shell_command = self.connection_protocol + " " + self.computer + " mkdir " + self.remote_benchmark_dir
      if self.gateway:
        gateway_command = self.connection_protocol + " " + self.gateway + " mkdir " + self.remote_benchmark_dir
	(status,output) = commands.getstatusoutput(gateway_command)
	if status !=0:
	  results.add_error(self, "Unable to create remote benchmark directory on gateway: " + output)
	  return (status,copied_files)
        shell_command = self.connection_protocol + " " + self.gateway + " " + shell_command
      (status, output) = commands.getstatusoutput(shell_command)
    if status != 0:
      results.add_error(self, "Unable to create remote benchmark directory: " + output)
      return (status, copied_files)
    for path in self.include_paths:
      if os.path.isdir(path):
        dirlist = os.listdir(path)
        for elem in dirlist:
          if not os.path.isdir(path + "/" + elem):
            elem = string.replace(path + "/" + elem,self.working_directory + "/",'')
            self.copy_include_files.append(elem)
    for file in self.copy_include_files:
      if file not in copied_files:
        if os.access(file,os.F_OK):
          status = self.make_subdirs(file)
          if status != 0:
            results.add_error(self, "Unable to create needed subdirectories for remote compilation")
            return (status, copied_files)
	  if self.gateway:
	    shell_command = self.transfer_protocol + " " + file + " " + self.gateway + ":" + self.remote_benchmark_dir + file
	    (status, output) = commands.getstatusoutput(shell_command)
            if status != 0:
              results.add_error(self, "Unable to copy files to gateway :" + output)
              return (status, copied_files)
	    shell_command = self.connection_protocol + " " + self.gateway + " " + self.transfer_protocol \
            + " " + self.remote_benchmark_dir + file + " " + self.computer + ":" + self.remote_benchmark_dir + file
	    (status, output) = commands.getstatusoutput(shell_command)
            if status != 0:
              results.add_error(self, "Unable to copy files from gateway to benchmark host :" + output)
              return (status, copied_files)
	  else:
            shell_command = self.transfer_protocol + " " + file + " " + self.computer + ":" + self.remote_benchmark_dir + file
	    (status, output) = commands.getstatusoutput(shell_command)
            if status != 0:
              results.add_error(self, "Unable to copy files needed for remote compilation to benchmark host :" + output)
              return (status, copied_files)
	  copied_files.append(file)
    return (status, copied_files)
    
  def make_subdirs(self, file):
    """
    This function makes the subdirectories on a remote computer needed 
    for the copy_include_files_to_host function. 
    """
    curpath = self.remote_benchmark_dir
    status = 0
    path = string.split(file,"/")
    fullpath = string.replace(file,path[-1],'')
    shell_command = self.connection_protocol + " " + self.computer + " cd " + self.remote_benchmark_dir + fullpath
    if self.gateway:
      shell_command = self.connection_protocol + " " + self.gateway + " " + shell_command
    (status, output) = commands.getstatusoutput(shell_command)
    if status !=0:
      path = path[0:-1]
      for elem in path:
        curpath = curpath + elem + "/"
        shell_command = self.connection_protocol + " " + self.computer + " cd " + curpath
        if self.gateway:
          shell_command = self.connection_protocol + " " + self.gateway + " " + shell_command
        (status, output) = commands.getstatusoutput(shell_command)
        if status != 0:
	  if self.gateway:
	    shell_command = self.connection_protocol + " " + self.gateway + " mkdir " + curpath
	    (status, output) = commands.getstatusoutput(shell_command)
          shell_command = self.connection_protocol + " " + self.computer + " mkdir " + curpath
          if self.gateway:
            shell_command = self.connection_protocol + " " + self.gateway + " " + shell_command
          (status, output) = commands.getstatusoutput(shell_command)
          if status != 0:
            break
    return status

  def tidy_up(self):
    """
    This function cleans up after the benchmark run.
    """
    status = 0
    if self.generated_driver and os.path.exists(self.driver_file):
      os.remove(self.driver_file)
    if self.remote_tidy_up and self.computer != os.uname()[1] and self.connection_protocol:
      shell_command =  self.connection_protocol + " " + self.computer + " rm -rf " + self.remote_benchmark_dir
      if self.gateway:
        gateway_command = self.connection_protocol + " " + self.gateway + " rm -rf " + self.remote_benchmark_dir
	(status,output) = commands.getstatusoutput(gateway_command) 
	shell_command = self.connection_protocol + " " + self.gateway + " " + shell_command
      (status, output) = commands.getstatusoutput(shell_command)
    else:
      executable = os.path.splitext(self.driver_file)[0]
      if os.path.exists(executable):
        os.remove(executable)
      da_file = executable + ".da"
      if os.path.exists(da_file):
        os.remove(da_file)
      
    return status

  def translate_filenames(self):
    """
    This function translates the paths of the files in the include_files list
    to match the remote benchmark directory.
    """
    self.copy_include_files = self.include_files
    temp_include_files = []
    for file in self.include_files:
      if os.path.isfile(file):
        if file[0]=='/':
          file = '.' + file
        else:
          file = './' + file
      temp_include_files.append(file)
    self.include_files = temp_include_files

  def output(self):
    """
    This function defines the result of a benchmark run, normally this is simply 
    the output of the driver program. This can be overridden by a class inheriting case
    for more advanced benchmarks.
    """
    return self.driver_output

  def __str__(self):
    """
    returns the objects class signature
    """
    if self.__class__.__doc__ is not None:
      s = string.strip(self.__class__.__doc__)
      first_line = string.split(s, '\n')[0]
      return first_line
    return "%s" % self.__class__

  def run(self, results):
    """
    This function runs the benchmark and returns the result and any errors encountered.
    """
    results.start(self)            
    answer = ()
    try:
      if self.driver_file is None:
        results.add_error(self, "No driver given")
      else:
        ok = 0
        try:
          if self.computer != os.uname()[1]:
            if not self.filenames_translated:
              self.translate_filenames()
            if not self.include_files_copied:
	      (status, copied_files) = self.copy_include_files_to_host(results)
              if status != 0:
                results.stop(self)
                return results
          results = self.do_experiment(self.driver_file, results)
          answer = self.output()
	  ok = 1
        except KeyboardInterrupt:
          results.abort(self)
	  ok = 0
	  raise
        except:
          results.add_error(self, self.exc_info_to_string(sys.exc_info()))
	  ok = 0
        try:
          status = self.tidy_up()
          if status != 0:
            results.add_error(self, "Unable to perform remote tidy up")
        except KeyboardInterrupt:
          results.abort(self)
	  ok = 0
	  raise
        except:
          results.add_error(self, self.exc_info_to_string(sys.exc_info()))
          ok = 0
	if ok:
          results.add_answer(self, answer)
    finally:
      results.stop(self)
      return results

  def debug(self):
    """
    runs do_experiment but ignores the result
    """
    ignore_results = self.do_experiment(self.driver_file, result())
    ignore_answer = self.output()
    status = self.tidy_up()

  def generate_cpu_time_driver(self):
    """
    This function generates a C++ driver for a functor.
    The functor must have a constructor and a member function primal, and
    possibly also a member function dual. The output is the time used by
    the primal minus that of the dual.
    """
    if self.computer != os.uname()[1]:
      self.translate_filenames()
      self.filenames_translated = 1
    if self.include_statements is None:
      self.include_statements = ""
      for file in self.include_files:
        if (file[0] == '<' and file[-1] == '>')\
          or (file[0] == '"' and file[-1] == '"'):   
          self.include_statements = self.include_statements + \
            '#include ' + file + '\n'
        else:
          self.include_statements = self.include_statements + \
            '#include "' + file + '"\n'
    if self.constructor_call is None:
      print >>sys.stderr, "self.constructor_call needed to generate a driver"
      sys.exit(1)

    self.time_factor = self.time_factors[self.time_unit]
    if self.dual_exists == None:
      self.dual_exists = self.check_dual() 
    if self.dual_exists:
      program = """\
%(include_statements)s
int main() {
  double primal_in_seconds = measure_primal(%(constructor_call)s);  
  double dual_in_seconds = measure_dual(%(constructor_call)s);  
  double t = (primal_in_seconds - dual_in_seconds) * %(time_factor)s;
  std::cout << t << std::endl;
  return 0;
}  
""" % self.__dict__ 
    else:
      program = """\
%(include_statements)s
int main() {
  double primal_in_seconds = measure_primal(%(constructor_call)s);  
  double t = primal_in_seconds * %(time_factor)s;
  std::cout << t << std::endl;
  return 0;
}  
""" % self.__dict__

    driver_file = self.working_directory + "/driver." + self.id + ".c++"
    f = open(driver_file, "w")
    f.write(program)
    f.close()
    self.generated_driver = 1
    if self.computer != os.uname()[1]:
      driver_file = string.split(driver_file,"/")[-1]
    return driver_file

#25.5.2003 \Jyrki
#This function should be split into two parts:
#generate_cache_miss_driver
#generate_page_fault_driver

  def generate_execution_mem_driver(self):
    """
    This function generates a C++ driver for a functor.
    The functor must have a constructor and a member function primal. The output is the page faults
    and cache misses. If self.papitype is single: a single run is measured, if it is quad: four runs
    are measured and averaged and if warmedup is specified: only the second of two runs are measured.
    The output of the program is the number of major page faults, the number of minor page faults, the
    number of L1 datacache misses and the number of L1 datacache acceses separated by semicolon
    """
    pythonpath = string.split(os.environ['PYTHONPATH'], ':') 
    for f in pythonpath:
      g = os.path.abspath(f)
      papi = os.path.abspath(g + '/Papibase')     
      if os.path.isdir(papi):
        self.include_paths.append(papi)
        list = os.listdir(papi)
        for h in list:
          i = os.path.abspath(papi + '/' + h)
          if os.path.isfile(i) and os.path.splitext(i)[1] == '.c++':
            self.include_files.append(i)
    if self.computer != os.uname()[1]:
      self.translate_filenames()
      self.filenames_translated = 1
    if self.include_statements is None:
      self.include_statements = ""
      for file in self.include_files:
        if (file[0] == '<' and file[-1] == '>')\
          or (file[0] == '"' and file[-1] == '"'):   
          self.include_statements = self.include_statements + \
            '#include ' + file + '\n'
        else:
          self.include_statements = self.include_statements + \
            '#include "' + file + '"\n'
    if self.constructor_call is None:
      print >>sys.stderr, "self.constructor_call needed to generate a driver"
      sys.exit(1)
    if self.papitype is None:
      self.papitype = "single"
    program = """\
%(include_statements)s
int main() {
  result r;
  int s = measure_page_primal_%(papitype)s(%(constructor_call)s,&r);  
  if (s) return -1;
  std::cout << r.majpgf << ";" << r.minpgf << ";" << r.L1cmiss << ";" << r.L1cacc << std::endl;
  return 0;
}  
""" % self.__dict__
    driver_file = self.working_directory + "/driver.mem." + self.id + ".c++"
    f = open(driver_file, "w")
    f.write(program)
    f.close()
    self.compiler_options.append("-lpapi")
    self.compiler_options.append("-lperfctr")
    if not self.papi_path:
      print "No papi path given. Aborting driver generation."
      return None    
    self.remote_include_paths.append(self.papi_path + "/include")
    self.compiler_options.append("-L" + self.papi_path + "/lib")
    self.generated_driver = 1
    if self.computer != os.uname()[1]:
      driver_file = string.split(driver_file,"/")[-1]
    return driver_file

  def generate_branch_misprediction_ratio_driver(self):
    """
    This function generates a C++ driver for a functor.
    The functor must have a constructor and a member function primal. The output is the branch missprediction
    ratio.
    """
    pythonpath = string.split(os.environ['PYTHONPATH'], ':') 
    for f in pythonpath:
      g = os.path.abspath(f)
      papi = os.path.abspath(g + '/Papibase')     
      if os.path.isdir(papi):
        self.include_paths.append(papi)
        list = os.listdir(papi)
        for h in list:
          i = os.path.abspath(papi + '/' + h)
          if os.path.isfile(i) and os.path.splitext(i)[1] == '.c++':
            self.include_files.append(i)
    if self.computer != os.uname()[1]:
      self.translate_filenames()
      self.filenames_translated = 1
    if self.include_statements is None:
      self.include_statements = ""
      for file in self.include_files:
        if (file[0] == '<' and file[-1] == '>')\
          or (file[0] == '"' and file[-1] == '"'):   
          self.include_statements = self.include_statements + \
            '#include ' + file + '\n'
        else:
          self.include_statements = self.include_statements + \
            '#include "' + file + '"\n'
    if self.constructor_call is None:
      print >>sys.stderr, "self.constructor_call needed to generate a driver"
      sys.exit(1)
    program = """\
%(include_statements)s
int main() {
  branchresult r;
  int s = measure_branch(%(constructor_call)s,&r);  
  if (s) return -1;
  std::cout << r.BRmisp << ";" << r.BR << std::endl;
  return 0;
}  
""" % self.__dict__

    driver_file = self.working_directory + "/driver.branch." + self.id + ".c++"
    f = open(driver_file, "w")
    f.write(program)
    f.close()
    self.compiler_options.append("-lpapi")
    self.compiler_options.append("-lperfctr")
    if not self.papi_path:
      print "No papi path given. Aborting driver generation."
      return None    
    self.remote_include_paths.append(self.papi_path + "/include")
    self.compiler_options.append("-L" + self.papi_path + "/lib")
    self.generated_driver = 1
    if self.computer != os.uname()[1]:
      driver_file = string.split(driver_file,"/")[-1]
    return driver_file

  def generate_function_call_count_driver(self):
    """
    This function generates a C++ driver for a functor.
    The functor must have a constructor and a member function primal. The function specified in
    self.counted_function is profiled. This function adds the g++ 3.2.2 switches needed to
    profile the program, if you use a compiler that needs other switches you have to 
    overwrite self.compiler_options "after" this function is called. 
    """
    if self.computer != os.uname()[1]:
      self.translate_filenames()
      self.filenames_translated = 1
    if self.include_statements is None:
      self.include_statements = ""
      for file in self.include_files:
        if (file[0] == '<' and file[-1] == '>')\
          or (file[0] == '"' and file[-1] == '"'):   
          self.include_statements = self.include_statements + \
            '#include ' + file + '\n'
        else:
          self.include_statements = self.include_statements + \
            '#include "' + file + '"\n'
    if self.constructor_call is None:
      print >>sys.stderr, "self.constructor_call needed to generate a driver"
      sys.exit(1)

    program = """\
%(include_statements)s
int main() {
  do_function(%(constructor_call)s);  
  return 0;
}  
""" % self.__dict__

    driver_file = self.working_directory + "/driver.func." + self.id + ".c++"
    f = open(driver_file, "w")
    f.write(program)
    f.close()
    self.compiler_options.append("-p -pg -fprofile-arcs")
    self.generated_driver = 1
    if self.computer != os.uname()[1]:
      driver_file = string.split(driver_file,"/")[-1]
    return driver_file

  def generate_compile_command(self, source, executable):
    """
    This function generates the compile command for the banchmark.
    In doing so the function tells the compiler to look for files 
    in the paths of include_paths and the remote_include_paths lists
    by use of the '-I' switch. 
    """
    f = string.replace(source, ' ', '\ ')
    f = string.replace(f, '"', '\"')
    f = string.replace(f, '<', '\<')
    f = string.replace(f, '>', '\>')
    f = string.replace(f, '&', '\&')
    f = string.replace(f, '|', '\|')
    compile_command = self.compiler + " "
    for option in self.compiler_options:
      compile_command = compile_command + option + " "
    if self.computer != os.uname()[1]:
      temp_paths = []
      for path in self.include_paths:
        path = self.remote_benchmark_dir[0:-1] + string.replace(path,self.working_directory,'')
        temp_paths.append(path)
      self.include_paths = temp_paths
    for file in self.include_paths:
      compile_command = compile_command + "-I" + file + " "
    for file in self.remote_include_paths:
      compile_command = compile_command + "-I" + file + " "
    compile_command = compile_command + "-o" + executable + " "
    compile_command = compile_command + f
    if self.computer != os.uname()[1] and self.connection_protocol:
      compile_command = self.connection_protocol + " " + self.computer \
        + " " + compile_command
      if self.gateway:
        compile_command = self.connection_protocol + " " + self.gateway + " " + compile_command
    return compile_command

  def check_dual(self):
    """
    This function checks whether the functor has a member function named dual or not.
    This check is done locally regardless of the computer selected to run the benchmark.
    If this check fails and you are sure you have a dual function it is most likely
    because your functor cannot compile correctly.
    """
    program = """
%(include_statements)s
template <typename functor>
void dual_exists(functor f) {
  f.dual();
}
int main() {
  dual_exists(%(constructor_call)s);
  return 0;
}
""" % self.__dict__
    filename = self.working_directory + "/dual." + self.id + ".c++"
    f = open(filename, "w")
    f.write(program)
    f.close()
    executable = os.path.splitext(filename)[0]
    f = string.replace(filename, ' ', '\ ')
    f = string.replace(f, '"', '\"')
    f = string.replace(f, '<', '\<')
    f = string.replace(f, '>', '\>')
    f = string.replace(f, '&', '\&')
    f = string.replace(f, '|', '\|')
    compile_command = self.compiler + " "
    for option in self.compiler_options:
      compile_command = compile_command + option + " "
    for file in self.include_paths:
      compile_command = compile_command + "-I" + file + " "
    compile_command = compile_command + "-o" + executable + " "
    shell_command = compile_command + f
    (status, output) = commands.getstatusoutput(shell_command)
    os.remove(filename)
    try:
      os.remove(executable)
    except:
      pass
    if status == 0:
      return 1
    else:
      return 0

  def generate_execute_command(self, executable):
    """
    This function generates the command used to execute the benchmark.
    If applicable this command include connections to a remote host and a
    gateway machine.
    """
    execute_command = executable
    if self.computer != os.uname()[1] and self.connection_protocol:
      execute_command = self.connection_protocol + " " + self.computer \
        + " " + execute_command
      if self.gateway:
        execute_command = self.connection_protocol + " " + self.gateway + " " + execute_command
    return execute_command

  def do_experiment(self, cpp_filename, results):
    """
    This function compiles the benchmark and executes it. It returns the result 
    and errors if any occured. 
    If the benchmark is to be run remotely, the function copies the driver to 
    the remote host before compiling it.
    """
    if self.computer != os.uname()[1]:
      if self.transfer_protocol and self.connection_protocol:
        if self.gateway:
          shell_command = self.transfer_protocol + " " + cpp_filename + " " + self.gateway + ":" + cpp_filename
	  (status, output) = commands.getstatusoutput(shell_command)
          if status != 0:
            results.add_error(self, "Unable to copy driver to gateway :" + output)
            return results
	  shell_command = self.connection_protocol + " " + self.gateway + " " + self.transfer_protocol \
          + " " + cpp_filename + " " + self.computer + ":" + self.remote_benchmark_dir + cpp_filename
	  (status, output) = commands.getstatusoutput(shell_command)
          if status != 0:
            results.add_error(self, "Unable to copy driver from gateway to benchmark host :" + output)
            return results
	  shell_command = self.connection_protocol + " "  + self.gateway + " rm -f " + cpp_filename
	  (status, output) = commands.getstatusoutput(shell_command)
          if status != 0:
            results.add_error(self, "Unable to remove driver from gateway :" + output)
        else:
          shell_command = self.transfer_protocol + " " + cpp_filename + " " + self.computer + ":" \
          + self.remote_benchmark_dir + cpp_filename
          (status, output) = commands.getstatusoutput(shell_command)
          if status != 0:
            results.add_error(self, "Unable to copy driver to benchmark host :" + output)
            return results
      else:
        result.add_error(self, "No connection protocol or transfer protocol given")
        return results
      cpp_filename = self.remote_benchmark_dir + cpp_filename
    executable = os.path.splitext(cpp_filename)[0]
    shell_command = self.compile_command
    if self.compile_command is None:
      shell_command = self.generate_compile_command(cpp_filename, executable)
    (status, output) = commands.getstatusoutput(shell_command)
    if status != 0:
      results.add_error(self, output)
      return results
    shell_command = self.execute_command
    if self.execute_command is None:
      shell_command = self.generate_execute_command(executable)
    (status, output) = commands.getstatusoutput(shell_command)
    if status != 0:
      results.add_error(self, output)
      return results
    
    if self.counted_function is not None:
      shell_command = "gprof -p -b " + executable + "|grep \" " + self.counted_function + "\""
      if self.computer != os.uname()[1]:
        if self.connection_protocol:
          shell_command = self.connection_protocol + " " + self.computer + " " + shell_command
          if self.gateway:
            shell_command = self.connection_protocol + " " + self.gateway + " " + shell_command
      (status,output) = commands.getstatusoutput(shell_command)
      if status != 0:
        results.add_error(self,"Could not profile: " + output)
        return results 
      
    self.driver_output = output
    return results
 
  def exc_info_to_string(self, error):
    """
    returns a string containing a traceback
    """
    return \
      string.strip(string.join(apply(traceback.format_exception, error), ''))

######################################################################
# suite
######################################################################

class suite:
  """
  A suite is a composite benchmark consisting of a number of cases or suites
  """

  def __init__(self, benchmark = ()):
    self.title = self.__str__()
    self.id = str(random.random())[2:]
    self.benchmarks = []
    self.add(benchmark)

  def __str__(self):
    """
    Returns the class signature
    """
    if self.__class__.__doc__ is not None:
      s = string.strip(self.__class__.__doc__)
      first_line = string.split(s, '\n')[0]
      return first_line
    return "%s" % self.__class__

  def add(self, benchmark):
    """
    Adds a benchmark or list of benchmarks to the list in this suite.
    """
    if isinstance(benchmark, types.ListType) or isinstance(benchmark, types.TupleType):
      for b in benchmark:
        self.benchmarks.append(b)
    else:
      self.benchmarks.append(benchmark)

  def run(self, results, copied_files = [], remote_benchmark_dir = ""):
    """
    This function runs either a list of suites or a list of benchmarks. 
    In the case of remote runs it assumes that all benchmarks are run on 
    the same host in the sense that only the host on which the first 
    benchmark is run gets cleaned up.
    """
    answers = []
    errors = []
    for b in self.benchmarks:
      if results.should_abort:
        break
      args = results.__getinitargs__()
      new_results = results.__class__(*args)      
      if isinstance(b, case):
        if b.computer != os.uname()[1]:
          if remote_benchmark_dir == "":
            remote_benchmark_dir = "./CPHSTL-Benchmarking-" + self.id + "/"
          b.remote_benchmark_dir = remote_benchmark_dir
          b.remote_tidy_up = 0
          if not b.filenames_translated:
            b.translate_filenames()
            b.filenames_translated = 1
	  (status, copied_files) = b.copy_include_files_to_host(new_results, copied_files)
	  b.include_files_copied = 1
        new_results = b.run(new_results)
        if new_results.should_abort:
	  result.should_abort = 1      
      else:
        new_results = b.run(new_results, copied_files)
	if new_results.should_abort:
	  result.should_abort = 1
      answers.extend(new_results.answers)
      errors.extend(new_results.errors)
    if len(answers) > 0:
      results.add_answer(self, answers)
    if len(errors) > 0:
      results.add_error(self, errors)
    status = self.remote_tidy_up(remote_benchmark_dir)
    if status != 0:
      print "Remote tidy up failed"
    return results

  def remote_tidy_up(self, dir):
    """
    This function removes the remote benchmark directory of the first 
    benchmark in the suite if this benchmark was run remotely.
    """
    benchmark = self.benchmarks[0]
    status = 0
    if isinstance(benchmark, case):
      if benchmark.computer != os.uname()[1]:
        if benchmark.connection_protocol:
          benchmark_dir = dir
          shell_command = benchmark.connection_protocol + " " + benchmark.computer + " rm -rf " + benchmark_dir
          if benchmark.gateway:
            gateway_command = benchmark.connection_protocol + " "  + benchmark.gateway + " rm -rf " + benchmark_dir
	    (status,output) = commands.getstatusoutput(gateway_command)
	    shell_command = benchmark.connection_protocol + " " + benchmark.gateway + " " + shell_command
          (status, output) = commands.getstatusoutput(shell_command)
        else:
          print "No connection protocol given"
    return status

  def debug(self):
    """
    Run the benchmarks without collecting errors in a result instance
    """
    for benchmark in self.benchmarks:
      benchmark.debug()

######################################################################
# curve_suite
######################################################################

class curve_suite(suite):
  """
  A suite specifying how a curve should be drawn by gnuplot
  Such a suite should only consist of benchmark cases.
  """

  def __init__(self, benchmark = ()):
    """
    initializes data structures and sets default options
    """
    suite.__init__(self, benchmark)
    self.style = 'linespoints'
    self.linetype = None
    self.linewidth = None
    self.pointtype = None
    self.pointsize = None

######################################################################
# plot_suite
######################################################################

class plot_suite(suite):
  """
  A suite specifying how a plot should be drawn by gnuplot
  Such a suite should only consist of curve suites.
  """

  def __init__(self, benchmark = ()):
    """
    initializes data structures and sets default options
    """
    suite.__init__(self, benchmark)
    self.output = os.path.abspath(os.getcwd()) + '/plot.' + self.id + '.ps' 
    self.terminal = 'postscript landscape noenhanced monochrome dashed defaultplex'
    self.key = 'right top Right noreverse samplen 4 spacing 1.25 title ""'
    self.style = 'linespoints'
    self.gnuplot_commands = """
set encoding iso_8859_1
set output "%(output)s"
set terminal %(terminal)s
set key %(key)s
set title '%(title)s'
set data style %(style)s
""" % self.__dict__

######################################################################
# Text UI; this is Steve's code
######################################################################

class _writeln_decorator:
    """
    Used to decorate file-like objects with a handy 'writeln' method
    """
    def __init__(self, stream):
        self.stream = stream

    def __getattr__(self, attr):
        return getattr(self.stream, attr)

    def writeln(self, *args):
        if args: 
          apply(self.write, args)
        self.write('\n') # text-mode streams translate to \r\n if needed

######################################################################
# text_result
######################################################################

class text_result(result):
  """
  A result class that can print formatted text results to a stream
  Used by text_runner.
  """
  separator1 = '=' * 70
  separator2 = '-' * 70

  def __init__(self, stream, verbosity):
    """
    Sets verbosity options.
    """
    result.__init__(self)
    self.stream = stream
    self.verbosity = verbosity
    self.show_all = verbosity > 1
    self.dots = verbosity == 1
    self.previous = None

  def __getinitargs__(self):
    """
    Returns stream and verbosity
    """
    return (self.stream, self.verbosity)

  def start(self, benchmark):
    """
    Outputs title if showall is on
    """
    result.start(self, benchmark)
    if self.show_all:
      self.stream.write(benchmark.title)
      self.stream.write(" ... ")

  def add_answer(self, benchmark, answer):
    """
    Adds answer to result and displays progress
    """
    result.add_answer(self, benchmark, answer)
    if isinstance(benchmark, case) and answer != ():
      if self.show_all:
        self.stream.writeln("Done")
      elif self.dots:
        self.stream.write('.')

  def add_error(self, benchmark, error):
    """
    Adds error and displays E
    """
    result.add_error(self, benchmark, error)
    if isinstance(benchmark, case) and benchmark != self.previous:
      if self.show_all:
        self.stream.writeln("ERROR")
      elif self.dots:
        self.stream.write('E')
      self.previous = benchmark

  def print_answers(self):
    """
    Prints answers and separators
    """
    if self.dots or self.show_all:
      self.stream.writeln()
    self.stream.writeln(self.separator1)
    for t in self.answers:
      self.print_answer(t, '')
    self.stream.writeln(self.separator2)

  def print_answer(self, tuple, indentation):
    """
    Prints benchmark results indenting suites 
    """
    b, s = tuple
    if isinstance(b, suite):
      self.stream.write(indentation)
      if b.title != '':
        self.stream.writeln(b.title)      
      indentation += '  '
      for i in s:
        self.print_answer(i, indentation)
    elif isinstance(b, case):
      self.stream.write(indentation)
      if b.title != '':
        self.stream.write("%s: " % b.title)
      if s == ():
        self.stream.writeln("")
      else:
        self.stream.writeln("%s" % str(s))        

  def print_errors(self):
    """
    Same as print_answers but prints errors
    """
    if self.dots or self.show_all:
      self.stream.writeln()
    self.stream.writeln(self.separator1)
    for t in self.errors:
      self.print_error(t)

  def print_error(self, tuple):
    """
    Same as print_answer but prints errors
    """
    b, s = tuple
    if isinstance(b, suite) and isinstance(s, types.ListType):
      for i in s:
        self.print_error(i)
    elif isinstance(b, case) \
      or (isinstance(b, suite) and isinstance(s, types.StringType)):
      if b.title != '':
        self.stream.writeln("ERROR: %s" % b.title)
      else:
        self.stream.writeln("ERROR: %s" % b.__str__())
      self.stream.writeln("%s" % str(s))        
      self.stream.writeln(self.separator2)

######################################################################
# gnuplot_result
######################################################################

class gnuplot_result(text_result):
  """
  A result class that can print the results in the form of a plot
  Used by gnuplot_runner.
  """
  def __init__(self, stream, verbosity):
    text_result.__init__(self, stream, verbosity)

  def print_answers(self):
    """
    Prints answers and separators.
    """
    if self.dots or self.show_all:
      self.stream.writeln()
    self.stream.writeln(self.separator1)
    for t in self.answers:
      self.print_answer(t)
    self.stream.writeln(self.separator2)

  def print_answer(self, tuple):
    """
    Prints status report, writes a gnuplotscript and executes gnuplot
    on the file.
    """
    b, s = tuple
    if isinstance(b, plot_suite):
      script_name = os.path.splitext(b.output)[0] + '.gp'
      script_file = open(script_name, 'w')
      script_stream = _writeln_decorator(script_file)
      script_stream.writeln(b.gnuplot_commands)
      self.draw_plots(s, script_stream)
      script_file.close()
      self.stream.writeln('Wrote a gnuplot script in file: ' + script_name)
      shell_command = "gnuplot " + script_name
      (status, output) = commands.getstatusoutput(shell_command)
      if status != 0:
        self.add_error(b, output)
        return
      self.stream.writeln('Wrote a plot in file: ' + b.output)
    elif isinstance(b, case):
      pass
    else:
      for i in s:
        self.print_answer(i)

  def draw_plots(self, plots, stream):
    """
    Writes a gnuplotscript with the data from plots to stream.
    """
    def cmpfunct(x, y):
      """ Reverse ordering """
      b1, l1 = x
      b2, l2 = y
      c1, t1 = l1[-1]
      c2, t2 = l2[-1]
      if t1 < t2:
        return 1
      elif t1 > t2:
        return -1
      else:
        return 0
        
    for b, s in plots:
      if not isinstance(b, curve_suite):
        self.add_error((b, "A plot suite can only contain curve suites"))
        return
    plots.sort(cmpfunct)
    stream.write('plot ')
    for b, s in plots:
      stream.write("'-' ")
      stream.write('title "%s" ' % b.title)
      if b.style is not None:
        stream.write('with ' + b.style)
        if b.linetype is not None:
          stream.write(" " + b.linetype)
        if b.linewidth is not None:
          stream.write(" " + b.linewidth)
        if b.pointtype is not None:
          stream.write(" " + b.pointtype)
        if b.pointsize is not None:
          stream.write(" " + b.pointsize)
      stream.write(", ")
    stream.writeln('0 notitle')
    for b, s in plots:
      for c, v in s:
        if v == ():
          stream.writeln("")
        else:
          for d in v:
            stream.write("%s " % d)
          stream.writeln("")
      stream.writeln("e")

######################################################################
# text_runner
######################################################################

class text_runner:
  """
  A benchmark runner class that displays the results in textual form
  It prints out the names of benchmarks as they are run, errors as they
  occur, and a summary of the results at the end of the benchmark run.
  """
  def __init__(self, stream = sys.stderr, verbosity = 1):
    self.stream = _writeln_decorator(stream)
    self.verbosity = verbosity

  def run(self, benchmark):
    """
    Run the given benchmark case or benchmark suite
    """
    results = text_result(self.stream, self.verbosity)
    start_time = time.time()
    results = benchmark.run(results)
    stop_time = time.time()
    time_taken = float(stop_time - start_time)
    number_of_errors = results.number_of_errors()
    if number_of_errors > 0:
      results.print_errors()
    results.print_answers()
    run = results.run_cases()
    ending = "s"
    if run == 1:
      ending = ""
    self.stream.writeln("Ran %d benchmark case%s in %.3f s" %
                       (run, ending, time_taken))
    if number_of_errors > 0:
      self.stream.write("Errors: ")
      self.stream.writeln("%d" % number_of_errors)
    else:
      self.stream.writeln("OK")
    return results

######################################################################
# gnuplot_runner
######################################################################

class gnuplot_runner(text_runner):
  """
  A benchmark runner class that displays the results in graphical form
  """
  def __init__(self, stream = sys.stderr, verbosity = 1):
    text_runner.__init__(self, stream, verbosity)

  def run(self, benchmark, result = gnuplot_result):
    """
    Run the given plot_suite
    """
    results = gnuplot_result(self.stream, self.verbosity)
    start_time = time.time()
    results = benchmark.run(results)
    stop_time = time.time()
    time_taken = float(stop_time - start_time)
    results.print_answers()
    number_of_errors = results.number_of_errors()
    if number_of_errors > 0:
      results.print_errors()
    run = results.run_cases()
    ending = "s"
    if run == 1:
      ending = ""
    self.stream.writeln("Ran %d benchmark case%s in %.3f s" %
                       (run, ending, time_taken))
    if number_of_errors > 0:
      self.stream.write("Errors: ")
      self.stream.writeln("%d" % number_of_errors)
    else:
      self.stream.writeln("OK")
    return results
    
##############################################################################
# Facilities for starting benchmarks from the command line
##############################################################################

class main:
  """
  A command-line program that can be used to run a given benchmark
  """
  USAGE = """\
Usage: %(program)s [options]

Optional options:
  -h, --help       Show this message
  -v, --verbose    Verbose output
  -q, --quiet      Minimal output
  --date           Output the date
  --runner=y       Use runner class y
  --task=x         Run benchmark instance x
  --version        Output the version of the tool

Examples:
  shell> %(program)s --version
  shell> %(program)s --task="partial_sort_case(0,1000000,\"int\",\"std\")"
  shell> %(program)s --task="combined_suite()" --runner="text_runner"
"""
  def __init__(self, argv = None, task = None, runner = text_runner):
    """
    Initializes datastructures and sets default options.
    Defaults:
    verbosity = 1
    main.runner = text_runner
    """
    if argv is None:
      argv = sys.argv
    self.verbosity = 1
    self.task = None
    self.runner = None
    self.program = os.path.basename(argv[0])
    self.parse_arguments(argv)
    if self.task is None:
      self.task = task
    if self.runner is None:
      self.runner = runner(verbosity = self.verbosity)
    self.run()

  def usage_and_exit(self, msg = None):
    """
    Prints the commandline arguments and exits
    """
    if msg: 
      print msg
    print self.USAGE % self.__dict__
    sys.exit(-1)

  def parse_arguments(self, argv):
    """
    Parses arguments possible arguments are:
    -h, --help:		displays the Usage message
    --date:		displays date
    -q, --quiet:	does not print progress
    -v, --verbose: 	prints everything
    --version:		prints version
    --task=t:   	sets main.task=t
    --runner=r:		sets main.runner=r	
    """
    import getopt
    try:
      options, args = getopt.getopt(argv[1:], 'hH', \
        ['date', 'help', 'verbose', 'version', 'quiet', 'task=', 'runner='])
      for opt, value in options:
        if opt in ('--date',):
	  today_tuple = time.localtime()
	  today_string = time.strftime("%d %B %Y %H:%M:%S", today_tuple)
	  print "%s (%s)" % (today_string, sys.argv[0])
        if opt in ('-h','-H','--help'):
          self.usage_and_exit()
        if opt in ('--quiet',):
          self.verbosity = 0
        if opt in ('--verbose',):
          self.verbosity = 2
        if opt in ('--version',):
	  date_tuple =  time.strptime(__date__, "%Y/%m/%d %H:%M:%S")             
	  release_date = time.strftime("%d %B %Y", date_tuple)             
	  print "%s %s (%s)" % (__tool__, __version__, release_date)
        if opt in ('--task',):
          self.task = eval(value)
        if opt in ('--runner',):
          self.runner = eval(value)
    except KeyboardInterrupt:         
      raise
    except getopt.error, msg:
      self.usage_and_exit(msg)

  def run(self):
    """
    Runs self.runner on the benchmark task
    """
    if self.task is not None:
      results = self.runner.run(self.task)
      sys.exit(results.number_of_errors())

if __name__ == "__main__":
  main()
